import git
import tempfile
import unittest
import os
import pathlib

import test_utils

import ocm
from concourse.steps import step_def
import concourse.model.traits.component_descriptor as component_descriptor


# make linter happy
if ocm.SchemaVersion.V2 == None:
    pass


class ComponentDescriptorStepTest(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()

        self.component_descriptor_trait = component_descriptor.ComponentDescriptorTrait(
            name='component_descriptor',
            variant_name='don\'t_care',
            raw_dict={
                'component_name': 'github.com/org/repo_name',
            },
        )
        self.component_descriptor_transformer = self.component_descriptor_trait.transformer()
        self.render_step = step_def('component_descriptor')

        self.main_repo = test_utils.repository()

        repo_dir = pathlib.Path(self.tmp_dir.name, self.main_repo.resource_name())
        repo_dir.mkdir()
        git.Repo.init(path=repo_dir)

        self.job_variant = test_utils.job(self.main_repo)

        self.job_variant._traits_dict = {'component_descriptor': self.component_descriptor_trait}

        for step in self.component_descriptor_transformer.inject_steps():
            self.job_variant._steps_dict[step.name] = step

        self.component_descriptor_step = self.job_variant.step(
            component_descriptor.DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME,
        )
        self.component_descriptor_step.add_input('version_path', 'version_path')

        self.old_cwd = os.getcwd()

    def tearDown(self):
        self.tmp_dir.cleanup()
        os.chdir(self.old_cwd)

    def test_render_and_compile(self):
        # as a smoke-test, just try to render
        step_snippet = self.render_step(
            job_step=self.job_variant.step('component_descriptor'),
            job_variant=self.job_variant,
            output_image_descriptors={},
            indent=0,
        )

        # try to compile (-> basic syntax check)
        return compile(step_snippet, 'component_descriptor.mako', 'exec')

    def test_smoke(self):
        os.chdir(self.tmp_dir.name)

        # fulfill runtime requirements

        # version file
        version_file_dir = pathlib.Path(self.tmp_dir.name, 'version_path')
        version_file_dir.mkdir()
        version_file = version_file_dir.joinpath('version')
        version_file.write_text('1.2.3')

        # component descriptor output directory
        component_descriptor_dir = pathlib.Path(self.tmp_dir.name, 'component_descriptor_dir')
        component_descriptor_dir.mkdir()

        compiled_step = self.test_render_and_compile()
        try:
            eval(compiled_step)
        except SystemExit:
            pass

        generated_component_descriptor = component_descriptor_dir.joinpath(
            'component_descriptor_v2'
        )
        self.assertTrue(generated_component_descriptor.is_file())
        # todo: parse and validate contents



import collections.abc
import dataclasses
import datetime
import logging
import requests
import time

import dacite

import ocm

import ccc.github
import ci.util
import cnudie.iter
import cnudie.retrieve
import cnudie.util
import delivery.jwt
import delivery.model as dm
import dso.model
import http_requests
import model
import model.base
import model.github


logger = logging.getLogger(__name__)


class DeliveryServiceRoutes:
    def __init__(self, base_url: str):
        self._base_url = base_url

    def auth(self):
        return ci.util.urljoin(
            self._base_url,
            'auth',
        )

    def auth_configs(self):
        return ci.util.urljoin(
            self._base_url,
            'auth',
            'configs',
        )

    def openid_configuration(self):
        '''
        endpoint according to OpenID provider configuration request
        https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderConfigurationRequest
        '''
        return ci.util.urljoin(
            self._base_url,
            '.well-known',
            'openid-configuration',
        )

    def component_descriptor(self):
        return ci.util.urljoin(
            self._base_url,
            'ocm',
            'component',
        )

    def greatest_component_versions(self):
        return ci.util.urljoin(
            self._base_url,
            'ocm',
            'component',
            'versions',
        )

    def component_responsibles(self):
        return ci.util.urljoin(
            self._base_url,
            'ocm',
            'component',
            'responsibles',
        )

    def _delivery(self, *suffix: collections.abc.Iterable[str]):
        return ci.util.urljoin(
            self._base_url,
            'delivery',
            *suffix,
        )

    def sprint_infos(self):
        return self._delivery('sprint-infos')

    def sprint_current(self):
        return self._delivery('sprint-infos', 'current')

    def artefact_metadata(self):
        return ci.util.urljoin(
            self._base_url,
            'artefacts',
            'metadata',
        )

    def artefact_metadata_query(self):
        return ci.util.urljoin(
            self._base_url,
            'artefacts',
            'metadata',
            'query',
        )

    def os_branches(self, os_id: str):
        return ci.util.urljoin(
            self._base_url,
            'os',
            os_id,
            'branches',
        )

    def components_metadata(self):
        return ci.util.urljoin(
            self._base_url,
            'components',
            'metadata',
        )


class DeliveryServiceClient:
    def __init__(
        self,
        routes: DeliveryServiceRoutes,
        github_cfgs: tuple[model.github.GithubConfig]=(),
        cfg_factory: model.ConfigFactory | None=None,
        auth_credentials: dm.GitHubAuthCredentials=None,
    ):
        '''
        Initialises a client which can be used to interact with the delivery-service.

        :param DeliveryServiceRoutes routes
            object which contains information of the base url of the desired instance of the
            delivery-service as well as the available routes
        :param tuple[GithubConfig] github_cfgs (optional):
            tuple of the available GitHub configurations which are used to authenticate against the
            delivery-service
        :param ConfigFactory cfg_factory (optional):
            the config factory is used to retrieve available GitHub configurations in case they are
            not provided anyways (i.e. can be safely omitted in case `github_cfgs` is specified)
        :param GitHubAuthCredentials auth_credentials (optional):
            object which contains credentials required for authentication against the
            delivery-service api
        '''
        self._routes = routes
        self.github_cfgs = github_cfgs
        self.cfg_factory = cfg_factory
        self.auth_credentials = auth_credentials

        self._bearer_token = None
        self._session = requests.sessions.Session()

    def _openid_configuration(self):
        '''
        response according to OpenID provider configuration response
        https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderConfigurationResponse
        '''
        res = self._session.get(
            url=self._routes.openid_configuration(),
            timeout=(4, 31),
        )

        res.raise_for_status()

        return res.json()

    def _openid_jwks(self):
        openid_configuration = self._openid_configuration()

        res = self._session.get(
            url=openid_configuration.get('jwks_uri'),
            timeout=(4, 31),
        )

        res.raise_for_status()

        return res.json()

    def _authenticate(self):
        if self._bearer_token and not delivery.jwt.is_jwt_token_expired(
            token=self._bearer_token,
        ):
            return

        if not self.auth_credentials:
            res = self._session.get(
                url=self._routes.auth_configs(),
                timeout=(4, 31),
            )

            res.raise_for_status()

            auth_configs = res.json()

            for auth_config in auth_configs:
                api_url = auth_config.get('api_url')

                try:
                    github_cfg = ccc.github.github_cfg_for_repo_url(
                        api_url=api_url,
                        cfg_factory=self.cfg_factory,
                        require_labels=(),
                        github_cfgs=self.github_cfgs,
                    )
                    break
                except model.base.ConfigElementNotFoundError:
                    continue
            else:
                logger.info('no valid credentials found - attempting anonymous-auth')
                return

            self.auth_credentials = dm.GitHubAuthCredentials(
                api_url=api_url,
                auth_token=github_cfg.credentials().auth_token(),
            )

        params = {
            'access_token': self.auth_credentials.auth_token,
            'api_url': self.auth_credentials.api_url,
        }

        res = self._session.get(
            url=self._routes.auth(),
            params=params,
            timeout=(4, 31),
        )

        if not res.ok:
            logger.warning(
                'authentication against delivery-service failed: '
                f'{res.status_code=} {res.reason=} {res.content=}'
            )

        res.raise_for_status()

        self._bearer_token = res.cookies.get(delivery.jwt.JWT_KEY)

        if not self._bearer_token:
            raise ValueError('delivery-service returned no bearer token upon authentication')

    def request(
        self,
        url: str,
        method: str='GET',
        headers: dict=None,
        **kwargs,
    ):
        self._authenticate()

        headers = headers or {}

        if self._bearer_token:
            headers = {
                'Authorization': f'Bearer {self._bearer_token}',
                **headers,
            }

        try:
            timeout = kwargs.pop('timeout')
        except KeyError:
            timeout = (4, 31)

        res = self._session.request(
            method=method,
            url=url,
            headers=headers,
            timeout=timeout,
            **kwargs,
        )

        return res

    def component_descriptor(
        self,
        name: str,
        version: str,
        ctx_repo_url: str=None,
        ocm_repo_url: str=None,
        version_filter: str | None=None,
        validation_mode: ocm.ValidationMode | None=None,
    ):
        params = {
            'component_name': name,
            'version': version,
        }
        ocm_repo_url = ocm_repo_url or ctx_repo_url
        if ocm_repo_url:
            params['ocm_repo_url'] = ocm_repo_url
        if version_filter is not None:
            params['version_filter'] = version_filter

        res = self.request(
            url=self._routes.component_descriptor(),
            params=params,
        )

        res.raise_for_status()

        return ocm.ComponentDescriptor.from_dict(
            res.json(),
            validation_mode=validation_mode,
        )

    def greatest_component_versions(
        self,
        component_name: str,
        max_versions: int=5,
        greatest_version: str=None,
        ocm_repo: ocm.OcmRepository=None,
        version_filter: str | None=None,
        start_date: datetime.date=None,
        end_date: datetime.date=None,
        timeout: tuple[float, float]=(4.0, 121.0),
    ):
        params = {
            'component_name': component_name,
            'max': max_versions,
        }
        if greatest_version:
            params['version'] = greatest_version
        if ocm_repo:
            if not isinstance(ocm_repo, ocm.OciOcmRepository):
                raise NotImplementedError(ocm_repo)
            params['ocm_repo_url'] = ocm_repo.oci_ref
        if version_filter is not None:
            params['version_filter'] = version_filter

        if start_date:
            params['start_date'] = start_date.isoformat()

        if end_date:
            params['end_date'] = end_date.isoformat()

        res = self.request(
            url=self._routes.greatest_component_versions(),
            params=params,
            timeout=timeout,
        )

        res.raise_for_status()

        return res.json()

    def update_metadata(
        self,
        data: collections.abc.Iterable[dso.model.ArtefactMetadata],
    ):
        headers = {
            'Content-Type': 'application/json',
        }

        data, headers = http_requests.encode_request(
            json={'entries': [
                dataclasses.asdict(
                    artefact_metadata,
                    dict_factory=ci.util.dict_to_json_factory,
                ) for artefact_metadata in data
            ]},
            headers=headers,
        )

        res = self.request(
            url=self._routes.artefact_metadata(),
            method='PUT',
            headers=headers,
            data=data,
            timeout=None,
        )

        res.raise_for_status()

    def delete_metadata(
        self,
        data: collections.abc.Iterable[dso.model.ArtefactMetadata],
    ):
        headers = {
            'Content-Type': 'application/json',
        }

        data, headers = http_requests.encode_request(
            json={'entries': [
                dataclasses.asdict(
                    artefact_metadata,
                    dict_factory=ci.util.dict_to_json_factory,
                ) for artefact_metadata in data
            ]},
            headers=headers,
        )

        res = self.request(
            url=self._routes.artefact_metadata(),
            method='DELETE',
            headers=headers,
            data=data,
            timeout=None,
        )

        res.raise_for_status()

    def component_responsibles(
        self,
        name: str=None,
        version: str=None,
        ctx_repo_url: str=None,
        ocm_repo_url: str=None,
        version_filter: str | None=None,
        component: ocm.Component | ocm.ComponentDescriptor=None,
        artifact: ocm.Artifact | str=None,
    ) -> tuple[dict, list[dm.Status]]:
        '''
        retrieves component-responsibles and optional status info.
        Status info can be used to communicate additional information, e.g. that responsible-label
        was malformed.
        Responsibles are returned as a list of typed user identities. Optionally, an artifact
        (or artifact name) may be passed. In this case, responsibles are filtered for the given
        resource definition. Note that an error will be raised if the given artifact does not declare
        a artifact of the given name.

        known types: githubUser, emailAddress, personalName
        example (single user entry): [
            {type: githubUser, username: <username>, source: <url>, github_hostname: <hostname>},
            {type: emailAddress, email: <email-addr>, source: <url>},
            {type: peronalName, firstName, lastName, source: <url>},
        ]
        '''

        if any((name, version, ocm_repo_url, ctx_repo_url)):
            if not all((name, version)):
                raise ValueError('either all or none of name and version must be set')
            elif component:
                raise ValueError('must pass either name, version (and ocm_repo_url) OR component')
        elif component and (component := cnudie.util.to_component(component)):
            name = component.name
            version = component.version
        else:
            raise ValueError('must either pass component or name, version (and ocm_repo_url)')

        url = self._routes.component_responsibles()

        params = {
            'component_name': name,
            'version': version,
        }
        if ocm_repo_url or ctx_repo_url:
            params['ocm_repo_url'] = ocm_repo_url or ctx_repo_url
        if version_filter is not None:
            params['version_filter'] = version_filter

        if artifact:
            if isinstance(artifact, ocm.Artifact):
                artifact_name = artifact.name
            else:
                artifact_name = artifact

            params['artifact_name'] = artifact_name

        if component:
            logger.info(f'{component.identity()=} {params=}')
        else:
            logger.info(f'{params=}')

        # wait for responsibles result
        # -> delivery service is waiting up to ~2 min for contributor statistics
        for _ in range(24):
            resp = self.request(
                url=url,
                params=params,
                timeout=(4, 121),
            )
            if resp.status_code != 202:
                break
            time.sleep(5)

        resp.raise_for_status()
        resp_json: dict = resp.json()

        responsibles = resp_json['responsibles']
        statuses_raw = resp_json.get('statuses', [])
        statuses = [
            dacite.from_dict(
                data_class=dm.Status,
                data=status_raw,
                config=dacite.Config(
                    cast=[
                        dm.StatusType,
                    ],
                ),
            )
            for status_raw in statuses_raw
        ]

        return responsibles, statuses

    def sprints(self) -> list[dm.Sprint]:
        resp = self.request(
            url=self._routes.sprint_infos(),
        )

        resp.raise_for_status()

        sprints_raw = resp.json()['sprints']

        return [
            dm.Sprint.from_dict(sprint_info)
            for sprint_info in sprints_raw
        ]

    def sprint_current(self, offset: int=0, before: datetime.date=None) -> dm.Sprint:
        extra_args = {}
        if before:
            if isinstance(before, datetime.date) or isinstance(before, datetime.date):
                extra_args['before'] = before.isoformat()
            else:
                extra_args['before'] = before

        resp = self.request(
            url=self._routes.sprint_current(),
            params={'offset': offset, **extra_args},
        )

        resp.raise_for_status()

        return dm.Sprint.from_dict(resp.json())

    def query_metadata(
        self,
        components: collections.abc.Iterable[ocm.Component]=(),
        artefacts: collections.abc.Iterable[dso.model.ComponentArtefactId]=(),
        type: dso.model.Datatype | tuple[dso.model.Datatype]=None,
        referenced_type: dso.model.Datatype | tuple[dso.model.Datatype]=None,
    ) -> tuple[dso.model.ArtefactMetadata]:
        '''
        Query artefact metadata from the delivery-db and parse it as `dso.model.ArtefactMetadata`.

        @param components:      component identities used for filtering; if no identities are
                                specified, no component filtering is done
        @param type:            datatype(s) used for filtering; if no datatype(s) is (are)
                                specified, no datatype filtering is done
        @param referenced_type: referenced datatype(s) used for filtering (only applies to artefact
                                metadata of type `rescorings`); if no datatype(s) is (are)
                                specified, no referenced datatype filtering is done
        '''
        if components and artefacts:
            raise ValueError('at most one of `artefacts` or `components` must be specified')

        params = dict()

        if type:
            params['type'] = type

        if referenced_type:
            params['referenced_type'] = referenced_type

        headers = {
            'Content-Type': 'application/json',
        }

        if components:
            entries = [
                {
                    'component_name': c.name,
                    'component_version': c.version,
                } for c in components
            ]
        else:
            entries = [
                dataclasses.asdict(artefact)
                for artefact in artefacts
            ]

        data, headers = http_requests.encode_request(
            json={'entries': entries},
            headers=headers,
        )

        res = self.request(
            url=self._routes.artefact_metadata_query(),
            method='POST',
            headers=headers,
            data=data,
            params=params,
            timeout=None,
        )

        res.raise_for_status()

        artefact_metadata_raw = res.json()

        return tuple(
            dso.model.ArtefactMetadata.from_dict(raw)
            for raw in artefact_metadata_raw
        )

    def os_release_infos(self, os_id: str, absent_ok=False) -> list[dm.OsReleaseInfo]:
        url = self._routes.os_branches(os_id=os_id)

        res = self.request(
            url=url,
        )

        if not absent_ok:
            res.raise_for_status()
        elif not res.ok:
            return None

        return [
            dm.OsReleaseInfo.from_dict(ri) for ri in res.json()
        ]

    def components_metadata(
        self,
        component_name: str,
        component_version: str=None,
        metadata_types: list[str]=[], # empty list returns _all_ metadata-types
        select: str=None, # either `greatestVersion` or `latestDate`
    ) -> list[dso.model.ArtefactMetadata]:
        '''
        returns a list of artifact-metadata for the given component

        One of 'select' and 'component_version' must be given. However, if 'select' is given as
        `greatestVersion`, 'version' must _not_ be given.
        '''
        url = self._routes.components_metadata()

        resp = self.request(
            url=url,
            params={
                'name': component_name,
                'version': component_version,
                'type': metadata_types,
                'select': select,
            },
            timeout=(4, 121),
        )

        resp.raise_for_status()

        return [
            dso.model.ArtefactMetadata.from_dict(raw)
            for raw in resp.json()
        ]

    def artefact_metadata_for_resource_node(
        self,
        resource_node: 'cnudie.iter.ResourceNode',
        types: list[str],
    ) -> collections.abc.Generator[dso.model.ArtefactMetadata, None, None]:
        '''Return an iterable that contains all stored `ArtefactMetadata` of the given type for the
        given resource node.

        For possible values for `type` see `dso.model.Datatype`.
        '''

        component = resource_node.component
        resource = resource_node.resource

        for component_metadata in self.components_metadata(
            component_name=component.name,
            metadata_types=types,
            component_version=component.version,
        ):
            if not component_metadata.artefact.component_name == component.name:
                continue
            if not component_metadata.artefact.artefact.artefact_name == resource.name:
                continue
            if not component_metadata.artefact.artefact.artefact_version == resource.version:
                continue

            yield component_metadata

    def metadata(
        self,
        component: cnudie.retrieve.ComponentName=None,
        artefact: str=None,
        node: cnudie.iter.Node=None,
        types: collections.abc.Iterable[str]=None,
    ) -> collections.abc.Generator[dso.model.ArtefactMetadata, None, None]:
        if component:
            component = cnudie.util.to_component_id(component)

        if types:
            types = tuple(types)

        if not (bool(component) ^ bool(node)):
            raise ValueError('exactly one of component, node must be passed')

        if node:
            component = node.component
            artefact = node.artefact

        if isinstance(artefact, ocm.Artifact):
            artefact_name = artefact.name
            artefact_version = artefact.version
        elif isinstance(artefact, str):
            artefact_name = artefact
            artefact_version = None

        for metadata in self.components_metadata(
            component_name=component.name,
            component_version=component.version,
            metadata_types=types,
        ):
            if not artefact:
                yield metadata
                continue

            # todo: also check for artefact-type + consider version is an optional attr
            #       + consider extra-id (keep it simple for now)
            artefact_id = metadata.artefact.artefact
            if artefact_name and artefact_id.artefact_name != artefact_name:
                continue
            if artefact_version and artefact_id.artefact_version != artefact_version:
                continue
            yield metadata


def _normalise_github_hostname(github_url: str):
    # hack: for github.com, we might get a different subdomain (api.github.com)
    github_hostname = ci.util.urlparse(github_url).hostname
    parts = github_hostname.strip('.').split('.')
    if parts[0] == 'api':
        parts = parts[1:]
    github_hostname = '.'.join(parts)

    return github_hostname.lower()


def github_users_from_responsibles(
    responsibles: collections.abc.Iterable[dict],
    github_url: str=None,
) -> collections.abc.Generator[dm.GithubUser, None, None]:
    '''
    returns a generator yielding all github-users from the given `responsibles`.
    use `DeliveryServiceClient.component_responsibles` to retrieve responsibles
    if github_url is given, only github-users on a matching github-host are returned.
    This is useful if the returned users should exist on a certain target github-instance.
    github_url is gracefully parsed down to relevant hostname. It is okay to pass-in, e.g.
    a repository- or github-user-URL for convenience.
    '''
    if github_url:
        target_github_hostname = _normalise_github_hostname(github_url)
    else:
        target_github_hostname = None

    for responsible in responsibles:
        for responsible_info in responsible:
            if not responsible_info['type'] == 'githubUser':
                continue
            username = responsible_info['username']
            github_hostname = _normalise_github_hostname(responsible_info['github_hostname'])

            if target_github_hostname and target_github_hostname != github_hostname:
                continue

            yield dm.GithubUser(username=username, github_hostname=github_hostname)



import logging

import ci.log
import concourse.replicator
import model.concourse

ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


## it may seem pointless to wrap replicate-pipelines - however, this will at least help
## linters to detect errors that would be missed if the call were inlined in mako
def replicate_pipelines(
    cfg_set,
    job_mapping: model.concourse.JobMapping,
    pipelines_not_to_delete: list,
):
    # prevent own replication pipeline from being removed
    def filter_own_pipeline(pipeline_name: str):
        return pipeline_name in pipelines_not_to_delete

    result = concourse.replicator.replicate_pipelines(
        cfg_set=cfg_set,
        job_mapping=job_mapping,
        unpause_pipelines=job_mapping.unpause_deployed_pipelines(),
        unpause_new_pipelines=job_mapping.unpause_new_pipelines(),
        expose_pipelines=job_mapping.expose_deployed_pipelines(),
        remove_pipelines_filter=filter_own_pipeline,
    )
    return result



import graphlib
import unittest

from concourse.model.base import ScriptType
from concourse.model.job import JobVariant

from concourse.model.step import PipelineStep


class JobVariantTest(unittest.TestCase):
    def examinee(self, name='Dont care'):
        variant = JobVariant(
            name='Dont care', raw_dict={}, resource_registry={}
        )
        # set steps dict, usually done by factory.
        variant._steps_dict = {}
        return variant

    def pipeline_step(self, name, is_synthetic=False,  **kwargs):
        return PipelineStep(
            name=name,
            is_synthetic=is_synthetic,
            script_type=ScriptType.BOURNE_SHELL,
            raw_dict=kwargs,
        )

    def test_step_ordering(self):
        examinee = self.examinee()
        examinee.add_step(self.pipeline_step(name='foo'))
        examinee.add_step(self.pipeline_step(name='bar', depends=['foo']))
        examinee.add_step(self.pipeline_step(name='baz', depends=['foo', 'bar']))

        ordered_steps = tuple(examinee.ordered_steps())

        assert len(ordered_steps) == 3

        first, second, third = ordered_steps

        assert len(first) == 1
        assert len(second) == 1
        assert len(third) == 1

        assert 'foo' in first
        assert 'bar' in second
        assert 'baz' in third

    def test_step_ordering_should_fail_on_circular_dependency(self):
        examinee = self.examinee()
        examinee.add_step(self.pipeline_step(name='foo', depends=['baz']))
        examinee.add_step(self.pipeline_step(name='bar', depends=['foo']))
        examinee.add_step(self.pipeline_step(name='baz', depends=['bar']))

        with self.assertRaises(graphlib.CycleError):
            examinee.ordered_steps()

    def test_step_ordering_should_resolve_dependencies_on_publish_step(self):
        examinee = self.examinee()
        examinee.add_step(self.pipeline_step(name='do_something'))
        # TODO: Don't hardcode step names
        examinee.add_step(self.pipeline_step(
            name='prepare',
            is_synthetic=True,
            depends=['foo', 'do_something']
        ))
        examinee.add_step(self.pipeline_step(name='publish', is_synthetic=True, depends=['prepare']))

        examinee.add_step(self.pipeline_step(name='foo', depends=['publish']))

        ordered_steps = tuple(examinee.ordered_steps())

        for step_tuple in ordered_steps:
            assert len(step_tuple) == 1

        ordered_steps = tuple(step_tuple[0] for step_tuple in ordered_steps)

        assert ordered_steps == ('do_something', 'prepare', 'publish', 'foo')

        examinee = self.examinee()
        examinee.add_step(self.pipeline_step(name='do_something'))
        # TODO: Don't hardcode step names
        examinee.add_step(self.pipeline_step(
            name='prepare',
            is_synthetic=True,
            depends=['foo', 'bar', 'baz', 'do_something']
        ))
        examinee.add_step(self.pipeline_step(name='publish', is_synthetic=True, depends=['prepare']))

        examinee.add_step(self.pipeline_step(name='foo', depends=('publish',)))
        examinee.add_step(self.pipeline_step(name='bar', depends=('foo',)))
        examinee.add_step(self.pipeline_step(name='baz', depends=('bar',)))

        ordered_steps = tuple(examinee.ordered_steps())

        # first two steps can run in parallel, followed by sequential remainder
        first_two_steps = ordered_steps[0]
        remainder_steps = ordered_steps[1:]

        assert len(first_two_steps) == 2
        assert set(first_two_steps) == {'do_something', 'prepare'}

        for step_tuple in remainder_steps:
            assert len(step_tuple) == 1

        remainder_steps = tuple(step_tuple[0] for step_tuple in remainder_steps)

        assert remainder_steps == ('publish', 'foo', 'bar', 'baz')

        examinee = self.examinee()
        examinee.add_step(self.pipeline_step(name='do_something'))

        # TODO: Don't hardcode step names
        examinee.add_step(self.pipeline_step(
            name='prepare',
            is_synthetic=True,
            depends=['foo', 'bar', 'do_something']
        ))
        examinee.add_step(
            self.pipeline_step(name='publish', is_synthetic=True, depends=['prepare'])
        )

        examinee.add_step(self.pipeline_step(name='foo', depends=['publish']))
        examinee.add_step(self.pipeline_step(name='bar', depends=['publish']))

        ordered_steps = tuple(examinee.ordered_steps())
        assert len(ordered_steps) == 3 # (do_sth, prepare), (publish), (foo, bar)

        first_two_steps = ordered_steps[0]
        second_step = ordered_steps[1]
        last_two_steps = ordered_steps[2]

        assert len(first_two_steps) == 2
        assert len(second_step) == 1
        assert len(last_two_steps) == 2

        assert set(first_two_steps) == {'do_something', 'prepare'}
        assert set(second_step) == {'publish'}
        assert set(last_two_steps) == {'foo', 'bar'}

    def test_step_ordering_should_fail_on_publish_step_if_synthetic_steps_in_cycle(self):
        examinee = self.examinee()
        examinee.add_step(self.pipeline_step(name='do_something'))
        # TODO: Don't hardcode step names
        examinee.add_step(self.pipeline_step(
            name='prepare',
            is_synthetic=True,
            depends=['foo', 'synthetic_foo'],
        ))
        examinee.add_step(self.pipeline_step(name='publish', is_synthetic=True, depends=['prepare']))

        examinee.add_step(self.pipeline_step(name='foo', depends=['publish']))
        examinee.add_step(self.pipeline_step(
            name='synthetic_foo',
            is_synthetic=True,
            depends=['publish'],
        ))

        with self.assertRaises(graphlib.CycleError):
            examinee.ordered_steps()

    def test_step_ordering_should_resolve_cycles_with_synthetic_steps(self):
        examinee = self.examinee()
        examinee.add_step(self.pipeline_step(name='do_something'))
        examinee.add_step(self.pipeline_step(
            name='synthetic_foo',
            is_synthetic=True,
            depends=['do_something', 'post_foo_step']
        ))
        examinee.add_step(self.pipeline_step(name='post_foo_step', depends=['synthetic_foo']))

        ordered_steps = tuple(examinee.ordered_steps())

        assert len(ordered_steps) == 2 # ('do_something', 'synthetic_foo'), then 'post_foo_step',

        first_two_steps, last_step = ordered_steps

        assert len(first_two_steps) == 2
        assert len(last_step) == 1

        assert set(first_two_steps) == {'do_something', 'synthetic_foo'}
        assert last_step[0] == 'post_foo_step'



import concurrent.futures
import datetime
from dateutil import parser
import functools
import logging
import shutil
import tarfile
import tempfile
import typing
import zipfile

import github3.exceptions
import github3.repos

import ccc.delivery
import ccc.github
import cnudie.iter
import cnudie.retrieve
import checkmarx.client
import checkmarx.model as model
import checkmarx.project
import checkmarx.tablefmt
import ci.util
import github.compliance.model
import model.checkmarx as cmmmodel
import reutil
import dso.labels
import dso.model

import ocm
import github.compliance.model as gcm

logger = logging.getLogger(__name__)


def scan_sources(
    component_descriptor: ocm.ComponentDescriptor,
    cx_client: checkmarx.client.CheckmarxClient,
    team_id: str,
    timeout_seconds: int,
    max_workers: int = 4, # only two scan will be run per user
    exclude_paths: typing.Sequence[str] = (),
    include_paths: typing.Sequence[str] = (),
    force: bool = False,
) -> model.FinishedScans:
    component_descriptor_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
        ocm_repository_lookup=cnudie.retrieve.ocm_repository_lookup(
            component_descriptor.component.current_ocm_repo,
        ),
    )

    components = tuple(
        component_node.component for component_node in cnudie.iter.iter(
            component=component_descriptor.component,
            lookup=component_descriptor_lookup,
            node_filter=cnudie.iter.Filter.components,
        )
    )

    # identify scan artifacts and collect them in a sequence
    artifacts_gen = _get_scan_artifacts_from_components(
        components=components,
    )

    return scan_artifacts(
        cx_client=cx_client,
        exclude_paths=exclude_paths,
        include_paths=include_paths,
        max_workers=max_workers,
        scan_artifacts=artifacts_gen,
        team_id=team_id,
        force=force,
        timeout_seconds=timeout_seconds,
    )


def _get_scan_artifacts_from_components(
    components: typing.List[ocm.Component],
) -> typing.Generator:
    for component in components:
        for source in component.sources:
            if source.type not in (
                ocm.ArtefactType.DIRECTORY_TREE,
                ocm.ArtefactType.GIT,
            ):
                source_name = f'{component.name}:{component.version}/{source.name}:{source.version}'
                logger.warning(f'skipping {source_name} with {source.type=}')
                continue

            if source.access.type is not ocm.AccessType.GITHUB:
                raise NotImplementedError

            cx_label = source.find_label(name=dso.labels.SourceScanLabel.name)
            if not cx_label:
                cx_label = component.find_label(name=dso.labels.SourceScanLabel.name)
            if cx_label:
                cx_label: dso.labels.SourceScanLabel = dso.labels.deserialise_label(label=cx_label)
                scan_policy = cx_label.value.policy
            else:
                scan_policy = dso.labels.ScanPolicy.SCAN

            if scan_policy is dso.labels.ScanPolicy.SKIP:
                logger.info('Note: No source scanning is configured according to ScanPolicy label')
                continue
            elif scan_policy is dso.labels.ScanPolicy.SCAN:
                pass # do scan
            else:
                raise NotImplementedError(scan_policy)

            source_project_label = source.find_label(
                name=dso.labels.SourceProjectLabel.name
            )

            if source_project_label:
                source_project_label = dso.labels.deserialise_label(label=source_project_label)
                source_project_label: dso.labels.SourceProjectLabel

                scan_artifact_name = source_project_label.value
            else:
                scan_artifact_name = None

            if not scan_artifact_name:
                scan_artifact_name = source.name

            logger.info(f'Using project name from cd-label {scan_artifact_name}')

            yield dso.model.ScanArtifact(
                source=source,
                name=scan_artifact_name,
                label=cx_label,
                component=component,
            )


def scan_artifacts(
    cx_client: checkmarx.client.CheckmarxClient,
    max_workers: int,
    scan_artifacts: typing.Tuple[dso.model.ScanArtifact],
    team_id: str,
    timeout_seconds: int,
    exclude_paths: typing.Sequence[str] = (),
    include_paths: typing.Sequence[str] = (),
    force: bool = False,
) -> model.FinishedScans:

    finished_scans = model.FinishedScans()

    artifacts = tuple(scan_artifacts)
    artifact_count = len(artifacts)
    finished = 0

    logger.info(f'will scan {artifact_count} artifacts')

    scan_func = functools.partial(
        scan_gh_artifact,
        exclude_paths=exclude_paths,
        include_paths=include_paths,
        force=force,
        timeout_seconds=timeout_seconds,
    )

    def init_scan(
        scan_artifact: dso.model.ScanArtifact,
    ) -> model.ScanResult:
        nonlocal cx_client
        nonlocal scan_func
        nonlocal team_id

        cx_project = checkmarx.project.init_checkmarx_project(
            checkmarx_client=cx_client,
            source_name=scan_artifact.name,
            team_id=team_id,
        )

        if scan_artifact.source.access.type is ocm.AccessType.GITHUB:
            return scan_func(
                cx_project=cx_project,
                scan_artifact=scan_artifact,
            )
        else:
            raise NotImplementedError

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for artifact in artifacts:
            futures.append(
                executor.submit(init_scan, artifact)
            )
        for future in concurrent.futures.as_completed(futures):
            scan_result = future.result()
            scan_result: model.ScanResult

            if scan_result.scan_succeeded:
                finished_scans.scans.append(scan_result)
            else:
                finished_scans.failed_scans.append(scan_result)

            finished += 1
            logger.info(f'remaining: {artifact_count - finished}')

    return finished_scans


def upload_and_scan_gh_artifact(
    artifact: dso.model.ScanArtifact,
    gh_repo: github3.repos.repo.Repository,
    cx_project: 'checkmarx.project.CheckmarxProject',
    path_filter_func: typing.Callable,
    source_commit_hash: str,
    force: bool,
    timeout_seconds: int,
) -> model.ScanResult:

    clogger = component_logger(artifact_name=artifact.name)

    last_scans = cx_project.get_last_scans()
    tm_start = datetime.datetime.now()
    if len(last_scans) < 1:
        clogger.info('no scans found in project history. '
                     f'Starting new scan hash={source_commit_hash}')
        scan_id = download_repo_and_create_scan(
            artifact_name=artifact.name,
            artifact_version=artifact.source.version,
            cx_project=cx_project,
            hash=source_commit_hash,
            repo=gh_repo,
            path_filter_func=path_filter_func,
        )
    else:
        last_scan = last_scans[0]
        scan_id = last_scan.id

        if cx_project.is_scan_finished(last_scan):
            clogger.info('no running scan found. Comparing hashes')
            if force or cx_project.is_scan_necessary(hash=source_commit_hash):
                clogger.info('current hash differs from remote hash in cx. '
                            f'New scan started for hash={source_commit_hash}')
                scan_id = download_repo_and_create_scan(
                    artifact_name=artifact.name,
                    artifact_version=artifact.source.version,
                    cx_project=cx_project,
                    hash=source_commit_hash,
                    repo=gh_repo,
                    path_filter_func=path_filter_func,
                )
            else:
                clogger.info('version/hash has already been scanned. Getting results of last scan')
        else:
            clogger.info(f'found a running scan id={last_scan.id}. Polling it')
            start_time = parser.parse(last_scan.dateAndTime.startedOn)
            # ignore if older than two hours, sometimes scans keep hanging and never end
            if (datetime.datetime.now() - start_time).total_seconds() > 2 * 60 * 60:
                clogger.info(f'running scan id={last_scan.id} found but older than two hours. '
                    f'Starting new scan hash={source_commit_hash}')
                scan_id = download_repo_and_create_scan(
                    artifact_name=artifact.name,
                    artifact_version=artifact.source.version,
                    cx_project=cx_project,
                    hash=source_commit_hash,
                    repo=gh_repo,
                    path_filter_func=path_filter_func,
                )

    scan_result = cx_project.poll_and_retrieve_scan(
        scan_id=scan_id,
        component=artifact.component,
        source=artifact.source,
        timeout_seconds=timeout_seconds,
    )
    duration = datetime.datetime.now() - tm_start
    clogger.info(f'Scan for component {artifact.name} took: {duration}')
    return scan_result


def _guess_commit_from_source(
    artifact_name: str,
    github_repo: github3.repos.repo.Repository,
    ref: str,
    commit_hash: str=None,
):
    def in_repo(commit_ish):
        try:
            return github_repo.ref(commit_ish).object.sha
        except github3.exceptions.NotFoundError:
            pass

        try:
            return github_repo.commit(commit_ish).sha
        except (github3.exceptions.UnprocessableEntity, github3.exceptions.NotFoundError):
            return None

    # first guess: look for commit hash if defined
    if commit_hash:
        commit = in_repo(commit_hash)
        if commit:
            return commit

    # second guess: check for ref like 'refs/heads/main'
    if ref.startswith('refs/'):
        gh_ref = ref[len('refs/'):] # trim 'refs/' because of github3 api
        commit = in_repo(gh_ref)
        if commit:
            return commit
    else:
        commit = in_repo(ref)
        if commit:
            return commit

    # third guess: branch
    try:
        return github_repo.branch(ref).commit.sha
    except github3.exceptions.NotFoundError:
        pass

    # still unknown commit-ish throw error
    raise RuntimeError(
        f'failed to guess on ref for {artifact_name=} with {ref=}'
    )


def scan_gh_artifact(
    cx_project: 'checkmarx.project.CheckmarxProject',
    scan_artifact: dso.model.ScanArtifact,
    timeout_seconds: int,
    exclude_paths: typing.Sequence[str] = (),
    include_paths: typing.Sequence[str] = (),
    force: bool = False,
) -> model.ScanResult:

    github_api = ccc.github.github_api_from_gh_access(access=scan_artifact.source.access)

    # access type has to be github thus we can call these methods
    try:
        gh_repo = github_api.repository(
            owner=scan_artifact.source.access.org_name(),
            repository=scan_artifact.source.access.repository_name(),
        )
    except Exception as e:
        logger.error(f'Failed to access Github repository, {scan_artifact.source.access.org_name()},'
            f'{scan_artifact.source.access.repository_name()}, on: '
            f'{scan_artifact.source.access.hostname()}')
        raise e

    try:
        commit_hash = _guess_commit_from_source(
            artifact_name=scan_artifact.name,
            commit_hash=scan_artifact.source.access.commit,
            github_repo=gh_repo,
            ref=scan_artifact.source.access.ref,
        )
    except github3.exceptions.NotFoundError as e:
        raise RuntimeError(e)

    if scan_artifact.label is not None:
        if scan_artifact.label.value.path_config is not None:
            path_config = scan_artifact.label.value.path_config
            include_paths = set((*include_paths, *path_config.include_paths))
            exclude_paths = set((*exclude_paths, *path_config.exclude_paths))

    # if the scan_artifact has no label we will implicitly scan everything
    # since all images have to specify a label in order to be scanned
    # only github access types can occour here without the label
    path_filter_func = reutil.re_filter(
        include_regexes=include_paths,
        exclude_regexes=exclude_paths,
    )
    return upload_and_scan_gh_artifact(
        artifact=scan_artifact,
        cx_project=cx_project,
        gh_repo=gh_repo,
        source_commit_hash=commit_hash,
        path_filter_func=path_filter_func,
        force=force,
        timeout_seconds=timeout_seconds,
    )


def greatest_severity(result: model.ScanResult) -> model.Severity | None:
    if not result.scan_statistic:
        raise RuntimeError('must only be called for successful scans')

    if result.scan_statistic.highSeverity > 0:
        return model.Severity.HIGH
    elif result.scan_statistic.mediumSeverity > 0:
        return model.Severity.MEDIUM
    elif result.scan_statistic.lowSeverity > 0:
        return model.Severity.LOW
    elif result.scan_statistic.infoSeverity > 0:
        return model.Severity.INFO
    else:
        return None


def checkmarx_severity_to_github_severity(severity: model.Severity) -> gcm.Severity:
    if not severity:
        raise RuntimeError('must only be called for successful scans')

    if severity in (
        model.Severity.HIGH,
        model.Severity.MEDIUM,
    ):
        return gcm.Severity.BLOCKER
    elif severity is model.Severity.LOW:
        return gcm.Severity.LOW
    elif severity is model.Severity.INFO:
        return None
    else:
        raise NotImplementedError(severity)


def print_scans(
    scans: model.FinishedScans,
    threshold: model.Severity,
    routes: checkmarx.client.CheckmarxRoutes,
):
    scans_above_threshold = [s for s in scans.scans
        if greatest_severity(s) and greatest_severity(s) >= threshold]
    scans_below_threshold = [s for s in scans.scans
        if greatest_severity(s) and greatest_severity(s) < threshold]

    if scans_above_threshold:
        print('\n')
        logger.info('critical scans above threshold')
        checkmarx.tablefmt.print_scan_result(
            scan_results=scans_above_threshold,
            routes=routes,
        )
    else:
        logger.info('no critical artifacts above threshold')

    if scans_below_threshold:
        print('\n')
        logger.info('clean scans below threshold')
        checkmarx.tablefmt.print_scan_result(
            scan_results=scans_below_threshold,
            routes=routes,
        )
    else:
        logger.info('no scans below threshold')

    if scans.failed_scans:
        print('\n')
        failed_artifacts_str = '\n'.join(
            (
                f'- {scan_result.artifact_name}:{scan_result.scanned_element.source.version}'
                for scan_result in scans.failed_scans
            )
        )
        logger.warning(f'failed scan artifacts:\n\n{failed_artifacts_str}\n')


def _download_and_zip_repo(
    clogger,
    path_filter_func: typing.Callable,
    ref: str,
    repo: github3.repos.repo.Repository,
    tmp_file
):
    files_to_scan = 0
    filtered_out_files = 0

    url = repo._build_url('tarball', ref, base_url=repo._api)
    with repo._get(url, stream=True) as r, \
        tarfile.open(fileobj=r.raw, mode='r|*') as tar, \
        zipfile.ZipFile(tmp_file, mode='w', compression=zipfile.ZIP_DEFLATED) as zip_file:

        r.raise_for_status()
        max_octets = 128 * 1024 * 1024  # 128 MiB

        # valid because first tar entry is root directory and has no trailing \
        path_offset = len(tar.next().name) + 1
        for tar_info in tar:
            if path_filter_func(arcname := tar_info.name[path_offset:]):
                if tar_info.isfile():
                    files_to_scan += 1
                    src = tar.extractfile(tar_info)
                    if tar_info.size <= max_octets:
                        zip_file.writestr(arcname, src.read())
                    else:
                        with tempfile.NamedTemporaryFile() as tmp:
                            shutil.copyfileobj(src, tmp)
                            zip_file.write(filename=tmp.name, arcname=arcname)
            else:
                filtered_out_files += 1

    clogger.info(f'{files_to_scan=}, {filtered_out_files=}')


def download_repo_and_create_scan(
    artifact_name: str,
    artifact_version: str,
    hash:str,
    cx_project: 'checkmarx.project.CheckmarxProject',
    path_filter_func: typing.Callable,
    repo: github3.repos.repo.Repository,
):
    clogger = component_logger(artifact_name=artifact_name)
    clogger.info('downloading sources from github')
    with tempfile.TemporaryFile() as tmp_file:
        _download_and_zip_repo(
            clogger=clogger,
            repo=repo,
            ref=hash,
            tmp_file=tmp_file,
            path_filter_func=path_filter_func,
        )
        tmp_file.seek(0)
        clogger.info('uploading sources to checkmarx')
        cx_project.upload_zip(file=tmp_file)

    cx_project.project_details.set_custom_field(
        attribute_key=model.CustomFieldKeys.HASH,
        value=hash,
    )
    cx_project.project_details.set_custom_field(
        attribute_key=model.CustomFieldKeys.COMPONENT_NAME,
        value=artifact_name,
    )
    cx_project.project_details.set_custom_field(
        attribute_key=model.CustomFieldKeys.VERSION,
        value=artifact_version,
    )
    cx_project.update_remote_project()

    scan_settings = model.ScanSettings(
        projectId=cx_project.project_details.id,
        comment=f'Scanning artifact name: {artifact_name}, version: {artifact_version}, '
            f'commit: {hash}'
    )
    scan_id = cx_project.start_scan(scan_settings)
    clogger.info(f'started scan id={scan_id} project id={cx_project.project_details.id}')

    return scan_id


@functools.lru_cache
def component_logger(artifact_name: str):
    return logging.getLogger(artifact_name)


def get_checkmarx_cfg(checkmarx_cfg_name: str) -> cmmmodel.CheckmarxConfig:
    cfg_fac = ci.util.ctx().cfg_factory()
    return cfg_fac.checkmarx(checkmarx_cfg_name)


@functools.lru_cache()
def create_checkmarx_client(checkmarx_cfg: cmmmodel.CheckmarxConfig):
    return checkmarx.client.CheckmarxClient(checkmarx_cfg)


def iter_artefact_metadata(
    results: typing.Iterable[model.ScanResult],
) -> typing.Generator[dso.model.ArtefactMetadata, None, None]:
    now = datetime.datetime.now()

    for result in results:
        artefact = github.compliance.model.artifact_from_node(result.scanned_element)
        artefact_ref = dso.model.component_artefact_id_from_ocm(
            component=result.scanned_element.component,
            artefact=artefact,
        )
        meta = dso.model.Metadata(
            datasource=dso.model.Datasource.CHECKMARX,
            type=dso.model.Datatype.CODECHECKS_AGGREGATED,
            creation_date=now,
            last_update=now,
        )
        codecheck = dso.model.CodecheckSummary(
            findings=dso.model.CodecheckFindings(
                high=result.scan_statistic.highSeverity,
                medium=result.scan_statistic.mediumSeverity,
                low=result.scan_statistic.lowSeverity,
                info=result.scan_statistic.infoSeverity,
            ),
            risk_rating=result.scan_response.scanRisk,
            risk_severity=result.scan_response.scanRiskSeverity,
            overview_url=result.overview_url,
            report_url=result.report_url,
        )

        yield dso.model.ArtefactMetadata(
            artefact=artefact_ref,
            meta=meta,
            data=codecheck,
        )



from copy import deepcopy
from itertools import chain
import graphlib

from ci.util import merge_dicts, not_none
from model.base import ModelValidationError
from concourse.model.step import PipelineStep

from concourse.model.base import (
        normalise_to_dict,
        ScriptType,
)
from concourse.validator import PipelineDefinitionValidator
from concourse.model.job import JobVariant
from concourse.model.pipeline import PipelineDefinition
from concourse.model.resources import (
    RepositoryConfig,
    ResourceRegistry,
)
from concourse.model.traits import TraitsFactory
from concourse.model.traits.meta import MetaTraitTransformer


def ensure_dict(d, allow_empty=True):
    if allow_empty and d is None:
        return {}
    if not allow_empty and not d:
        raise ModelValidationError('a non-empty dict is required')
    if not isinstance(d, dict):
        raise ModelValidationError('a dict is required')
    return d


class RawPipelineDefinitionDescriptor:
    '''
    Container type holding a single (raw) pipeline definition and metadata.
    Basic value validation is done in the c'tor.
    '''

    def __init__(
        self,
        name,
        base_definition,
        jobs,
        exception=None,
        template='default'
    ):
        self.name = not_none(name)
        self.base_definition = ensure_dict(base_definition, allow_empty=True)
        self.jobs = ensure_dict(jobs, allow_empty=False)
        self.template = not_none(template)
        self.exception = exception


class DefinitionFactory:
    '''
    Creates `PipelineDefinition` instances from "raw" `PipelineDefinitionDescriptor`s.

    "Raw" definitions are associative arrays of a certain structure. They are read from
    declaring components and are typically maintained by (human) component owners.

    Definitions feature an (optional) two-level inheritance hierarchy: A base definition and
    an arbitrary set of jobs. At least one variant is required. Jobs represent concrete
    single job definitions. Attributes defined in a base definition are inherited into each
    variant. Jobs may overwrite inherited attributes.
    '''

    def __init__(
        self,
        raw_definition_descriptor: RawPipelineDefinitionDescriptor,
        cfg_set,
    ):
        self.raw_definition_descriptor = not_none(raw_definition_descriptor)
        self.cfg_set = cfg_set

    def create_pipeline_definition(self) -> PipelineDefinition:
        merged_variants_dict = self._create_variants_dict(self.raw_definition_descriptor)

        resource_registry = ResourceRegistry()
        jobs = {}

        for variant_name, variant_dict in merged_variants_dict.items():
            variant = self._create_variant(
                raw_dict=variant_dict,
                variant_name=variant_name,
                resource_registry=resource_registry,
            )
            self._apply_traits(variant)

            # collect repositories
            for repo in chain(variant._repos_dict.values(), variant._publish_repos_dict.values()):
                if repo in resource_registry:
                    existing_repo = resource_registry.resource(repo)
                    # hack: patch-in should-trigger (the proper way to implement this
                    # would be to separate (effective) resource-definitions from additional
                    # resource-specialisations as contained in jobs
                    existing_repo._trigger |= repo.should_trigger()
                else:
                    resource_registry.add_resource(deepcopy(repo), discard_duplicates=False)

            jobs[variant_name] = variant

        pipeline_definition = PipelineDefinition()
        pipeline_definition._variants_dict = jobs
        pipeline_definition._resource_registry = resource_registry

        validator = PipelineDefinitionValidator(pipeline_definition=pipeline_definition)
        validator.validate()

        return pipeline_definition

    def _create_variants_dict(self, raw_definition_descriptor):
        variants_dict = normalise_to_dict(deepcopy(raw_definition_descriptor.jobs))

        base_dict = deepcopy(raw_definition_descriptor.base_definition)

        merged_variants = {}
        for variant_name, variant_args in variants_dict.items():
            # optimisation: if there are no variant-specific arguments, we do not need to merge
            if variant_args:
                merged_variants[variant_name] =  merge_dicts(base_dict, variant_args)
            else:
                merged_variants[variant_name] = deepcopy(base_dict)

        return merged_variants

    def _create_variant(self, raw_dict, variant_name, resource_registry) -> JobVariant:
        variant = JobVariant(
            name=variant_name,
            raw_dict=raw_dict,
            resource_registry=resource_registry
        )

        # build steps
        variant._steps_dict = self._create_build_steps(raw_dict)

        # traits
        variant._traits_dict = self._create_traits(raw_dict, variant_name)

        self._create_repos(variant, raw_dict)
        self._inject_publish_repos(variant)

        return variant

    def _apply_traits(self, pipeline_def):
        transformers = [trait.transformer() for trait in pipeline_def._traits_dict.values()]
        transformers_dict = {t.name: t for t in transformers}
        transformer_names = set(transformers_dict.keys())

        for transformer in transformers:
            if not set(transformer.dependencies()).issubset(transformer_names):
                missing = set(transformer.dependencies()) - transformer_names
                raise ModelValidationError(
                    f'{pipeline_def}: trait requires missing traits: ' + ', '.join(missing)
                )

        # order transformers according to dependencies
        toposorter = graphlib.TopologicalSorter()
        for transformer in transformers:
            dependencies = transformer.order_dependencies() & transformer_names
            toposorter.add(transformer.name, *dependencies)

        ordered_transformers = [
            transformers_dict[name] for name in toposorter.static_order()
        ]

        # hardcode meta trait transformer
        ordered_transformers.append(MetaTraitTransformer())

        # inject new steps
        for transformer in ordered_transformers:
            for step in transformer.inject_steps():
                pipeline_def.add_step(step)

        # do remaining processing
        for transformer in ordered_transformers:
            transformer.process_pipeline_args(pipeline_def)

    def _create_traits(self, raw_dict, variant_name):
        if 'traits' not in raw_dict:
            raw_dict['traits'] = {}

        if 'options' not in raw_dict['traits']:
            raw_dict['traits']['options'] = {}
        if 'notifications' not in raw_dict['traits']:
            raw_dict['traits']['notifications'] = {}

        traits_args = normalise_to_dict(raw_dict['traits'])
        traits_dict = {
                name: TraitsFactory.create(
                    name=name,
                    variant_name=variant_name,
                    args_dict=args if args else {},
                    cfg_set=self.cfg_set,
                )
                for name, args in traits_args.items()
        }
        return traits_dict

    def _create_build_steps(self, raw_dict):
        steps_dict = {}
        if 'steps' not in raw_dict:
            return steps_dict
        elif not raw_dict['steps']:
            return {}

        for stepname, step in raw_dict['steps'].items():
            if step is None:
                raw_dict['steps'][stepname] = {}

        steps_dict = {
            n: self._create_build_step(name=n, step_dict=sd) for n,sd in raw_dict['steps'].items()
        }
        return steps_dict

    def _create_build_step(self, name: str, step_dict: dict):
        return PipelineStep(
            name=name,
            is_synthetic=False,
            raw_dict=step_dict,
            script_type=ScriptType.BOURNE_SHELL,
        )

    def _create_repos(self, pipeline_def: JobVariant, raw_dict):
        pipeline_def._repos_dict = {}
        if 'repo' in raw_dict:
            # special case: repo singleton (will vanish once we mv definitions into component-repos)
            repo_dict = raw_dict['repo']
            name = 'source' if 'name' not in repo_dict else repo_dict['name']
            pipeline_def._repos_dict[name] = RepositoryConfig(
                raw_dict=repo_dict,
                logical_name=name,
                qualifier=None,
                is_main_repo=True
            )
            pipeline_def._main_repository_name = name
        if 'repos' in raw_dict:
            for repo_dict in raw_dict['repos']:
                if not 'cfg_name' in repo_dict:
                    github_cfg = self.cfg_set.github()
                else:
                    github_cfg = self.cfg_set.github(repo_dict['cfg_name'])

                hostname = github_cfg.hostname()
                repo_dict['hostname'] = hostname

                pipeline_def._repos_dict.update({
                    repo_dict['name']: RepositoryConfig(
                        logical_name=repo_dict['name'],
                        raw_dict=repo_dict, is_main_repo=False
                    )
                })

    def _inject_publish_repos(self, pipeline_def):
        # synthesise "put-repositories"
        pipeline_def._publish_repos_dict = {}
        for step in filter(lambda step: step.publish_repository_names(), pipeline_def.steps()):
            for repo_name in step.publish_repository_names():
                # we need to clone the existing repository configuration
                source_repo = pipeline_def._repos_dict[repo_name]
                publish_repo = RepositoryConfig(
                    raw_dict=dict(source_repo.raw),
                    logical_name=repo_name,
                    qualifier='output',
                    is_main_repo=source_repo.is_main_repo(),
                )
                pipeline_def._publish_repos_dict[repo_name] = publish_repo



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import enum
import urllib.parse

import dacite

from concourse.model.job import AbortObsoleteJobs
from model.base import ModelBase


class EventBase(ModelBase):
    def __init__(self, raw_dict, delivery, hostname):
        super().__init__(raw_dict)
        self._delivery = delivery
        self._hostname = hostname

    def repository(self):
        return Repository(self.raw['repository'])

    def delivery(self) -> str:
        '''return the id of the GitHub delivery that triggered this event
        '''
        return self._delivery

    def hostname(self) -> str:
        '''return the hostname of the GitHub instance that sent this event, e.g.: 'github.com'
        '''
        return self._hostname


class Repository(ModelBase):
    def github_host(self):
        return urllib.parse.urlparse(self.repository_url()).hostname

    def repository_url(self):
        url = self.raw['clone_url']
        if url.endswith('.git'):
            return url[:-4] # remove '.git' suffix
        return url

    def repository_path(self):
        return self.raw['full_name']


class RefType(enum.Enum):
    REPOSITORY = 'repository'
    BRANCH = 'branch'
    TAG = 'tag'


class CreateEvent(EventBase):
    def ref_type(self):
        return RefType(self.raw['ref_type'])

    def ref(self):
        '''
        @return: the ref's name or None if ref_type is repository
        '''
        return self.raw.get('ref', None)


class PushEvent(EventBase):
    def ref(self):
        return self.raw['ref']

    def modified_paths(self):
        # for now, only take head-commit into account
        # --> this could lead to missed updates
        head_commit = self.raw.get('head_commit', None)
        if not head_commit:
            return ()
        yield from head_commit.get('modified', ())

    def commit_message(self):
        # not all push events have a head_commit (e.g. push events sent on branch deletion)
        if head_commit := self.raw.get('head_commit'):
            return head_commit.get('message')
        return None

    def is_forced_push(self):
        return self.raw['forced']

    def previous_ref(self):
        return self.raw['before']


class PullRequestAction(enum.Enum):
    ASSIGNED = 'assigned'
    UNASSIGNED = 'unassigned'
    REVIEW_REQUESTED = 'review_requested'
    REVIEW_REQUEST_REMOVED = 'review_request_removed'
    LABELED = 'labeled'
    UNLABELED = 'unlabeled'
    OPENED = 'opened'
    EDITED = 'edited'
    CLOSED = 'closed'
    REOPENED = 'reopened'
    SYNCHRONIZE = 'synchronize'
    READY_FOR_REVIEW = 'ready_for_review'
    CONVERTED_TO_DRAFT = 'converted_to_draft'
    AUTO_MERGE_ENABLED = 'auto_merge_enabled'
    UNKNOWN = 'unknown'


class PullRequestEvent(EventBase):
    def action(self) -> PullRequestAction:
        action = self.raw['action']

        if not action in [
            pr_action.value
            for pr_action
            in PullRequestAction
        ]:
            return PullRequestAction.UNKNOWN

        return PullRequestAction(action)

    def number(self):
        '''
        the PR number
        '''
        return self.raw['number']

    def pr_id(self):
        '''
        the PR id

        Note: This is _not_ the same as its number.
        '''
        return self.raw['pull_request']['id']

    def label_names(self):
        '''Return the labels attached to the pull request as per the event received.
        '''
        return [
            label.get('name') for label in self.raw.get('pull_request').get('labels')
        ]

    def label(self) -> str:
        '''Return the Label that was set for `pull_request.labeled` events.
        '''
        if not self.action() is PullRequestAction.LABELED:
            raise RuntimeError(
                "Retrieving the triggering label is only valid for PRs with action 'labeled'"
            )
        return self.raw['label']['name']

    def fork(self):
        if (pr_info := self.raw.get('pull_request')):
            if (head_info := pr_info.get('head')):
                if (fork := head_info.get('fork')):
                    return fork
        raise RuntimeError(
            'Unable to determine whether this PR-event was sent for a PR created from a fork. '
        )

    def head_repository(self):
        return Repository(self.raw['pull_request']['head']['repo'])

    def sender(self):
        '''
        the user who performed the event
        '''
        return self.raw['sender']

    def head_commit(self):
        '''the SHA-hash of the head-commit of the PR

        returns None if unable to determine
        '''
        if (pr_info := self.raw.get('pull_request')):
            if (head_info := pr_info.get('head')):
                if (sha := head_info.get('sha')):
                    return sha
        return None

    def head_ref(self):
        '''the ref of the head-commit of the PR

        returns None if unable to determine
        '''
        if (pr_info := self.raw.get('pull_request')):
            if (head_info := pr_info.get('head')):
                if (ref := head_info.get('ref')):
                    return ref
        return None


@dataclasses.dataclass
class Pipeline:
    pipeline_name: str
    target_team: str
    effective_definition: dict


@dataclasses.dataclass
class AbortConfig:
    abort_obsolete_jobs: AbortObsoleteJobs

    @staticmethod
    def from_dict(d: dict):
        return dacite.from_dict(
            data_class=AbortConfig,
            data=d,
            config=dacite.Config(cast=[AbortObsoleteJobs]),
        )



import dataclasses
import datetime
import dateutil.parser
import enum
import os
import re
import typing

import dacite
import pytimeparse

import ci.util
import model.base as mb


class PolicyType(enum.Enum):
    MAX_AGE = 'max_age'


class RotationMethod(enum.Enum):
    MANUAL = 'manual'
    AUTOMATED = 'automated'


@dataclasses.dataclass
class CfgPolicy:
    '''
    re-usable policies that are intended to be assigned to configuration elements
    '''
    name: str
    max_age: typing.Optional[str]
    type: PolicyType = PolicyType.MAX_AGE
    rotation_method: RotationMethod = RotationMethod.MANUAL
    grace_period: str | None = '1d'
    delete_after_period: str | None = '7d'
    comment: typing.Optional[str] = None

    def check(
            self,
            last_update: datetime.date,
            honour_grace_period: bool = False,
        ) -> bool:
        '''returns `True` if policy is fulfilled, `False` if it is violated

        hard-coded to only allow PolicyType.MAX_AGE for now
        '''
        if self.max_age is None:
            return True # max_age being None means there is no expiry date

        max_age_seconds = pytimeparse.parse(self.max_age)

        if self.grace_period and honour_grace_period:
            grace_period = datetime.timedelta(seconds=pytimeparse.parse(self.grace_period))
        else:
            grace_period = datetime.timedelta(seconds=0)

        today = datetime.datetime.now()
        latest_required_update_date = (
            last_update + datetime.timedelta(seconds=max_age_seconds) + grace_period
        )

        return today < latest_required_update_date


@dataclasses.dataclass
class CfgTarget:
    type: str
    name: str

    def matches(
        self,
        element: typing.Union[str, mb.NamedModelElement],
        type: typing.Optional[str]=None,
    ) -> bool:
        '''
        checks whether the given model element matches this target. Model elements can be
        given in two forms:

        1. as only argument (-> element). In this case, `type` must not be passed (determined
           from the passed element)
        2. as two arguments (element being the element's name, type being the element's type's name)
           in the second form, both arguments must be strings
        '''
        if isinstance(element, mb.NamedModelElement):
            if type is not None:
                raise ValueError('if element is `NamedModelElement`, type must not be passed')

            element_name = element._name
            element_type = element._type_name
        else:
            if not type and element:
                raise ValueError('both element name and type name must be passed')

            element_name = ci.util.check_type(element, str)
            element_type = ci.util.check_type(type, str)

        type_matches = re.fullmatch(self.type, element_type)

        if not type_matches:
            return False

        name_matches = re.fullmatch(self.name, element_name)

        if not name_matches:
            return False

        return True # both type and name match


@dataclasses.dataclass
class CfgRule:
    '''
    configuration rules that map cfg policies to configuration elements
    '''
    targets: list[CfgTarget]
    policy: str

    def  matches(
        self,
        element: typing.Union[str, mb.NamedModelElement],
        type: typing.Optional[str]=None,
    ):
        for t in self.targets:
            if t.matches(element=element, type=type):
                return True

        return False


class CfgResponsibleType(enum.Enum):
    GITHUB = 'github'
    EMAIL = 'email'


# hash used to determine compliance rate for a responsible
@dataclasses.dataclass(unsafe_hash=True)
class CfgResponsible:
    name: str
    type: CfgResponsibleType


@dataclasses.dataclass
class CfgQueueEntry:
    target: CfgTarget
    deleteAfter: str
    secretId: dict

    def to_be_deleted(
        self,
        timestamp: datetime.datetime,
    ) -> bool:
        return timestamp > dateutil.parser.isoparse(self.deleteAfter)


@dataclasses.dataclass
class CfgStatus:
    target: CfgTarget
    credential_update_timestamp: str

    def  matches(
        self,
        element: typing.Union[str, mb.NamedModelElement],
        type: typing.Optional[str]=None,
    ):
        return self.target.matches(element=element, type=type)


@dataclasses.dataclass
class CfgResponsibleMapping:
    targets: list[CfgTarget]
    responsibles: list[CfgResponsible]

    def matches(
        self,
        element: typing.Union[str, mb.NamedModelElement],
        type: typing.Optional[str]=None,
    ):
        for t in self.targets:
            if t.matches(element=element, type=type):
                return True

        return False


@dataclasses.dataclass
class CfgMetadata:
    '''
    stores full cfg-element metadata from a cfg-dir
    '''
    policies: tuple[CfgPolicy]
    rules: tuple[CfgRule]
    responsibles: tuple[CfgResponsibleMapping]
    statuses: list[CfgStatus]
    queue: list[CfgQueueEntry]


class CfgElementPolicyViolation(enum.Enum):
    ASSIGNED_RULE_REFERS_TO_UNDEFINED_POLICY = 'assignedRuleRefersToUndefinedPolicy'
    CREDENTIALS_OUTDATED = 'credentialsOutdated'
    NO_RESPONSIBLE = 'noResponsible'
    NO_RULE = 'noRule'
    NO_STATUS = 'noStatus'


@dataclasses.dataclass(frozen=True)
class CfgStatusEvaluationResult:
    fullyCompliant: bool
    hasResponsible: bool
    hasRule: bool
    assignedRuleRefersToUndefinedPolicy: bool
    hasStatus: bool
    requiresStatus: typing.Optional[bool]
    credentialsOutdated: typing.Optional[bool]
    nonCompliantReasons: typing.List[CfgElementPolicyViolation]


@dataclasses.dataclass
class CfgResponsibleSummary:
    url: str
    responsible: CfgResponsible
    compliantElementsCount: int = 0
    noncompliantElementsCount: int = 0


@dataclasses.dataclass
class CfgStorageSummary:
    '''
    represents a compliance summary for a cfg_storage (url)
    '''
    url: str
    noRuleAssigned: list
    noStatus: list
    assignedRuleRefersToUndefinedPolicy: list
    noResponsibleAssigned: list
    credentialsOutdated: list
    credentialsNotOutdated: list
    fullyCompliant: list
    compliantElementsCount: int = 0
    noncompliantElementsCount: int = 0


@dataclasses.dataclass
class CfgElementStatusReport:
    '''
    represents the current status of a configuration element

    primarily targeted for creating reports for human consumers
    '''
    element_storage: str # e.g. a github-url - not intended to be machine-readable
    element_type: str
    element_name: str

    policy: typing.Optional[CfgPolicy]
    rule: typing.Optional[CfgRule]
    responsible: typing.Optional[CfgResponsibleMapping]
    status: typing.Optional[CfgStatus]

    @property
    def name(self) -> str:
        return f'{self.element_storage}/{self.element_type}/{self.element_name}'


cfg_policies_fname = 'config_policies.yaml'
cfg_responsibles_fname = 'config_responsibles.yaml'
cfg_status_fname = 'config_status.yaml'
cfg_queue_fname = 'config_queue.yaml'
container_registry_fname = 'container_registry.yaml'


def _parse_cfg_policies_file(path: str):
    raw = ci.util.parse_yaml_file(path=path)

    # document expected structure
    return {
        'policies': raw['policies'],
        'rules': raw['rules'],
    }


def _parse_cfg_responsibles_file(path: str):
    raw = ci.util.parse_yaml_file(path=path)

    # document expected structure
    return {
        'responsibles': raw['responsibles'],
    }


def _parse_cfg_status_file(path: str):
    raw = ci.util.parse_yaml_file(path=path)

    # document expected structure
    return {
        'config_status': raw['config_status'],
    }


def _parse_cfg_queue_file(path: str):
    raw = ci.util.parse_yaml_file(path=path)

    # document expected structure
    return {
        'rotation_queue': raw['rotation_queue'],
    }


def cfg_policies(policies: list[dict]) -> list[CfgPolicy]:
    if isinstance(policies, dict):
        policies = policies['policies']

    if not policies:
        policies = []

    return [
        dacite.from_dict(
            data_class=CfgPolicy,
            data=policy_dict,
            config=dacite.Config(
                cast=[PolicyType, RotationMethod],
            )
        ) for policy_dict in policies
    ]


def cfg_rules(rules: list[dict]) -> list[CfgRule]:
    if isinstance(rules, dict):
        rules = rules['rules']

    if not rules:
        rules = []

    return [
        dacite.from_dict(
            data_class=CfgRule,
            data=rule_dict,
        ) for rule_dict in rules
    ]


def cfg_responsibles(responsibles: list[dict]) -> list[CfgResponsibleMapping]:
    if isinstance(responsibles, dict):
        responsibles = responsibles['responsibles']

    if not responsibles:
        responsibles = []

    return [
        dacite.from_dict(
            data_class=CfgResponsibleMapping,
            data=responsible_dict,
            config=dacite.Config(
                cast=[CfgResponsibleType],
            )
        ) for responsible_dict in responsibles
    ]


def cfg_status(status: list[dict]) -> list[CfgStatus]:
    if isinstance(status, dict):
        status = status['config_status']

    if not status:
        status = []

    return [
        dacite.from_dict(
            data_class=CfgStatus,
            data=status_dict,
        ) for status_dict in status
    ]


def cfg_queue(queue: list[dict]) -> list[CfgQueueEntry]:
    if isinstance(queue, dict):
        queue = queue['rotation_queue']

    if not queue:
        queue = []

    return [
        dacite.from_dict(
            data_class=CfgQueueEntry,
            data=queue_dict,
        ) for queue_dict in queue
    ]


def cfg_metadata_from_cfg_dir(cfg_dir: str):
    policies = _parse_cfg_policies_file(os.path.join(cfg_dir, cfg_policies_fname))
    responsibles = _parse_cfg_responsibles_file(os.path.join(cfg_dir, cfg_responsibles_fname))
    statuses = _parse_cfg_status_file(os.path.join(cfg_dir, cfg_status_fname))
    queue = _parse_cfg_queue_file(os.path.join(cfg_dir, cfg_queue_fname))

    return CfgMetadata(
        policies=tuple(cfg_policies(policies['policies'])),
        rules=tuple(cfg_rules(policies['rules'])),
        responsibles=tuple(cfg_responsibles(responsibles['responsibles'])),
        statuses=list(cfg_status(statuses['config_status'])),
        queue=list(cfg_queue(queue['rotation_queue'])),
    )


class ValidationError(Exception):
    pass



import collections.abc
import logging
import os
import subprocess
import tempfile
import time
import traceback

import ocm
import github3.exceptions
import github3.repos.repo

import ccc.github
import ci.util
import cnudie.util
import concourse.model.traits.update_component_deps as ucd
import concourse.steps.component_descriptor_util as cdu
import dockerutil
import dso.labels
import github.util as gu
import gitutil
import model.container_registry as cr
import release_notes.fetch as release_notes_fetch
import version


logger = logging.getLogger('step.update_component_deps')


def current_product_descriptor():
    component_descriptor_file_path = cdu.component_descriptor_path(
        schema_version=ocm.SchemaVersion.V2,
    )

    # cd is supplied via component-descriptor file. Parse and return
    if os.path.isfile(component_descriptor_file_path):
        return ocm.ComponentDescriptor.from_dict(
            component_descriptor_dict=ci.util.parse_yaml_file(component_descriptor_file_path,)
        )
    else:
        raise RuntimeError(f'did not find component-descriptor at {component_descriptor_file_path=}')


def current_component():
    return current_product_descriptor().component


def close_obsolete_pull_requests(
    upgrade_pull_requests,
    reference_component: ocm.Component,
):
    obsolete_upgrade_requests = [
        pr for pr in upgrade_pull_requests
        if pr.pull_request.state == 'open' and pr.is_obsolete(
            reference_component=reference_component,
        )
    ]

    for obsolete_request in obsolete_upgrade_requests:
        obsolete_request.purge()


def upgrade_pr_exists(
    component_reference: ocm.ComponentReference,
    component_version: str,
    upgrade_requests: collections.abc.Iterable[gu.UpgradePullRequest],
    request_filter: collections.abc.Callable[[gu.UpgradePullRequest], bool] = lambda rq: True,
) -> gu.UpgradePullRequest | None:
    if any(
        (matching_rq := upgrade_rq).target_matches(
            reference=component_reference,
            reference_version=component_version,
        ) and request_filter(upgrade_rq)
        for upgrade_rq in upgrade_requests
    ):
        return matching_rq
    return None


def greatest_component_version(
    component_name,
    version_lookup,
    ignore_prerelease_versions,
) -> str | None:
    versions = version_lookup(component_name)
    if not versions:
        return None

    greatest_version = version.greatest_version(
        versions=versions,
        ignore_prerelease_versions=ignore_prerelease_versions,
    )

    if not greatest_version:
        return None

    return greatest_version


def greatest_component_version_with_matching_minor(
    component_name,
    version_lookup,
    reference_version,
    ignore_prerelease_versions,
) -> str | None:
    versions = version_lookup(component_name)
    if not versions:
        return None

    greatest_version = version.greatest_version_with_matching_minor(
        reference_version=reference_version,
        versions=versions,
        ignore_prerelease_versions=ignore_prerelease_versions,
    )

    if not greatest_version:
        return None

    return greatest_version


def latest_component_version_from_upstream(
    component_name: str,
    upstream_component_name: str,
    ocm_lookup,
    version_lookup,
    ignore_prerelease_versions: bool=False,
):
    upstream_component_version = greatest_component_version(
        component_name=upstream_component_name,
        version_lookup=version_lookup,
        ignore_prerelease_versions=ignore_prerelease_versions,
    )

    if not upstream_component_version:
        raise RuntimeError(
            f'did not find any versions for {upstream_component_name=}'
        )

    upstream_component_descriptor = ocm_lookup(
        ocm.ComponentIdentity(
            name=upstream_component_name,
            version=upstream_component_version,
        ),
    )

    upstream_component = upstream_component_descriptor.component
    for component_ref in upstream_component.componentReferences:
        # TODO: Validate that component_name is unique
        if component_ref.componentName == component_name:
            return component_ref.version


def determine_reference_versions(
    component_name: str,
    reference_version: str,
    version_lookup,
    ocm_lookup,
    upstream_component_name: str=None,
    upstream_update_policy: ucd.UpstreamUpdatePolicy=ucd.UpstreamUpdatePolicy.STRICTLY_FOLLOW,
    ignore_prerelease_versions: bool=False,
) -> collections.abc.Sequence[str]:
    if upstream_component_name is None:
        # no upstream component defined - look for greatest released version
        latest_component_version = greatest_component_version(
            component_name=component_name,
            version_lookup=version_lookup,
            ignore_prerelease_versions=ignore_prerelease_versions,
        )
        if not latest_component_version:
            raise RuntimeError(
                f'did not find any versions of {component_name=}'
            )

        return (
            latest_component_version,
        )

    version_candidate = latest_component_version_from_upstream(
        component_name=component_name,
        upstream_component_name=upstream_component_name,
        version_lookup=version_lookup,
        ocm_lookup=ocm_lookup,
        ignore_prerelease_versions=ignore_prerelease_versions,
    )

    if upstream_update_policy is ucd.UpstreamUpdatePolicy.STRICTLY_FOLLOW:
        return (version_candidate,)

    elif upstream_update_policy is ucd.UpstreamUpdatePolicy.ACCEPT_HOTFIXES:
        hotfix_candidate = greatest_component_version_with_matching_minor(
            component_name=component_name,
            version_lookup=version_lookup,
            reference_version=reference_version,
            ignore_prerelease_versions=ignore_prerelease_versions,
        )
        if hotfix_candidate == version_candidate:
            return (version_candidate,)
        else:
            return (hotfix_candidate, version_candidate)

    else:
        raise NotImplementedError


def greatest_references(
    references: collections.abc.Iterable[ocm.ComponentReference],
) -> collections.abc.Iterable[ocm.ComponentReference]:
    '''
    yields the component references from the specified iterable of ComponentReference that
    have the greatest version (grouped by component name).
    Id est: if the sequence contains exactly one version of each contained component name,
    the sequence is returned unchanged.
    '''
    references = tuple(references)
    names = {r.name for r in references}

    for name in names:
        matching_refs = [r for r in references if r.name == name]
        if len(matching_refs) == 1:
            # in case reference name was unique, do not bother sorting
            # (this also works around issues from non-semver versions)
            yield matching_refs[0]
        else:
            # there might be multiple component versions of the same name
            # --> use the greatest version in that case
            matching_refs.sort(key=lambda r: version.parse_to_semver(r.version))
            # greates version comes last
            yield matching_refs[-1]


def deserialise_extra_component_references(
    extra_crefs_label: dso.labels.ExtraComponentReferencesLabel,
) -> collections.abc.Generator[ocm.ComponentReference, None, None]:
    for extra_cref in extra_crefs_label.value:
        extra_cref_id = extra_cref.component_reference

        yield ocm.ComponentReference(
            name=extra_cref_id.name,
            componentName=extra_cref_id.name,
            version=extra_cref_id.version,
        )


def determine_upgrade_prs(
    upstream_component_name: str,
    upstream_update_policy: ucd.UpstreamUpdatePolicy,
    upgrade_pull_requests: collections.abc.Iterable[gu.UpgradePullRequest],
    version_lookup,
    ocm_lookup,
    ignore_prerelease_versions=False,
) -> collections.abc.Iterable[tuple[
    ocm.ComponentReference, ocm.ComponentReference, str
]]:
    component = current_component()
    component_references = component.componentReferences

    # don't use the deserialisation within `cnudie.iter.iter` here to avoid unnecessary lookups of
    # component references and keep `ComponentReference` instances (!= `Component` instances)
    if extra_crefs_label := component.find_label(
        name=dso.labels.ExtraComponentReferencesLabel.name,
    ):
        extra_crefs_label = dso.labels.deserialise_label(extra_crefs_label)
        extra_component_references = list(deserialise_extra_component_references(extra_crefs_label))
        component_references = component_references + extra_component_references

    for greatest_component_reference in greatest_references(
        references=component_references,
    ):
        versions_to_consider = determine_reference_versions(
            component_name=greatest_component_reference.componentName,
            reference_version=greatest_component_reference.version,
            upstream_component_name=upstream_component_name,
            upstream_update_policy=upstream_update_policy,
            version_lookup=version_lookup,
            ocm_lookup=ocm_lookup,
            ignore_prerelease_versions=ignore_prerelease_versions,
        )
        if versions_to_consider:
            logger.info(
                f"Found possible version(s) to up- or downgrade to: '{versions_to_consider}' for "
                f'{greatest_component_reference.componentName=}'
            )
        else:
            logger.warning(
                f'No component versions found for {greatest_component_reference.componentName=}'
            )
        for candidate_version in versions_to_consider:
            # we might have found 'None' as version to consider.
            if not candidate_version:
                continue

            candidate_version_semver = version.parse_to_semver(candidate_version)
            reference_version_semver = version.parse_to_semver(greatest_component_reference.version)

            logger.info(f'{candidate_version=}, ours: {greatest_component_reference}')

            if candidate_version_semver <= reference_version_semver:
                downgrade_pr = True
                # downgrades are permitted iff the version is tracking a _dependency_ of another
                # component and we are to follow strictly
                if (
                    candidate_version_semver == reference_version_semver or
                    not upstream_component_name
                    or upstream_update_policy is not ucd.UpstreamUpdatePolicy.STRICTLY_FOLLOW
                ):
                    logger.info(
                        f'skipping (outdated) {greatest_component_reference=}; '
                        f'our {greatest_component_reference.version=}, '
                        f'found: {candidate_version=}'
                    )
                    continue
            else:
                downgrade_pr = False

            if not downgrade_pr and (matching_pr := upgrade_pr_exists(
                component_reference=greatest_component_reference,
                component_version=candidate_version,
                upgrade_requests=upgrade_pull_requests,
                request_filter=lambda rq: not rq.is_downgrade(),
            )):
                logger.info(
                    'skipping upgrade (PR already exists): '
                    f'{greatest_component_reference=} '
                    f'to {candidate_version=} ({matching_pr.pull_request.html_url})'
                )
                continue
            elif downgrade_pr and (matching_pr := upgrade_pr_exists(
                component_reference=greatest_component_reference,
                component_version=candidate_version,
                upgrade_requests=upgrade_pull_requests,
                request_filter=lambda rq: rq.is_downgrade(),
            )):
                logger.info(
                    'skipping downgrade (PR already exists): '
                    f'{greatest_component_reference=} '
                    f'to {candidate_version=} ({matching_pr.pull_request.html_url})'
                )
                continue
            else:
                yield(greatest_component_reference, candidate_version)


def _import_release_notes(
    component: ocm.Component,
    to_version: str,
    pull_request_util,
    version_lookup,
    component_descriptor_lookup,
):
    if not component.sources:
        logger.warning(
            f'''
            {component.name=}:{component.version=} has no sources; skipping release-notes-import
            '''
        )
        return None

    main_source = cnudie.util.determine_main_source_for_component(component)
    github_cfg = ccc.github.github_cfg_for_repo_url(main_source.access.repoUrl)
    org_name = main_source.access.org_name()
    repository_name = main_source.access.repository_name()

    release_notes = create_release_notes(
        from_component=component,
        from_github_cfg=github_cfg,
        from_repo_owner=org_name,
        from_repo_name=repository_name,
        to_version=to_version,
        version_lookup=version_lookup,
        component_descriptor_lookup=component_descriptor_lookup,
    )

    if not release_notes:
        release_notes = pull_request_util.retrieve_pr_template_text()

    return release_notes


def create_upgrade_pr(
    component: ocm.Component,
    from_ref: ocm.ComponentReference,
    to_ref: ocm.ComponentReference,
    to_version: str,
    pull_request_util: gu.PullRequestUtil,
    upgrade_script_path,
    upgrade_script_relpath,
    githubrepobranch: gu.GitHubRepoBranch,
    repo_dir,
    github_cfg_name,
    merge_policy: ucd.MergePolicy,
    merge_method: ucd.MergeMethod,
    version_lookup,
    component_descriptor_lookup,
    after_merge_callback=None,
    container_image:str=None,
    pullrequest_body_suffix: str=None,
) -> gu.UpgradePullRequest:
    if container_image:
        dockerutil.launch_dockerd_if_not_running()

    ls_repo = pull_request_util.repository

    from_component_descriptor = component_descriptor_lookup(
        ocm.ComponentIdentity(
            name=from_ref.componentName,
            version=from_ref.version,
        ),
        absent_ok=False,
    )
    from_component = from_component_descriptor.component

    # prepare env for upgrade script and after-merge-callback
    cmd_env = os.environ.copy()
    # TODO: Handle upgrades for types other than 'component'
    cmd_env['DEPENDENCY_TYPE'] = 'component'
    cmd_env['DEPENDENCY_NAME'] = to_ref.componentName
    cmd_env['LOCAL_DEPENDENCY_NAME'] = to_ref.name
    cmd_env['DEPENDENCY_VERSION'] = to_version
    if container_image:
        cmd_env['REPO_DIR'] = (repo_dir_in_container := '/mnt/main_repo')
    else:
        cmd_env['REPO_DIR'] = repo_dir
    cmd_env['GITHUB_CFG_NAME'] = github_cfg_name
    cmd_env['CTX_REPO_URL'] = component.current_ocm_repo.baseUrl

    if not container_image:
        # create upgrade diff
        subprocess.run(
            [str(upgrade_script_path)],
            check=True,
            env=cmd_env
        )
    else:
        # run check-script in container
        oci_registry_cfg = cr.find_config(image_reference=container_image)
        if oci_registry_cfg:
            docker_cfg_dir = tempfile.TemporaryDirectory()
            dockerutil.mk_docker_cfg_dir(
                cfg={'auths': oci_registry_cfg.as_docker_auths()},
                cfg_dir=docker_cfg_dir.name,
                exist_ok=True,
            )
        else:
            docker_cfg_dir = None

        upgrade_script_path_in_container = os.path.join(
            repo_dir_in_container,
            upgrade_script_relpath,
        )

        docker_argv = dockerutil.docker_run_argv(
            image_reference=container_image,
            argv=(
                upgrade_script_path_in_container,
            ),
            env=cmd_env,
            mounts={
                repo_dir: repo_dir_in_container,
            },
            cfg_dir=docker_cfg_dir.name,
        )

        logger.info(f'will run: ${docker_argv=}')

        try:
            subprocess.run(
                docker_argv,
                check=True,
            )
        finally:
            if docker_cfg_dir:
                docker_cfg_dir.cleanup()

    from_version = from_ref.version
    commit_message = f'Upgrade {to_ref.name}\n\nfrom {from_version} to {to_version}'

    upgrade_branch_name = push_upgrade_commit(
        ls_repo=ls_repo,
        commit_message=commit_message,
        githubrepobranch=githubrepobranch,
        repo_dir=repo_dir,
    )
    # branch was created. Cleanup if something fails
    try:
        release_notes = _import_release_notes(
            component=from_component,
            to_version=to_version,
            pull_request_util=pull_request_util,
            version_lookup=version_lookup,
            component_descriptor_lookup=component_descriptor_lookup,
        )
    except Exception:
        logger.warning('failed to retrieve release-notes')
        traceback.print_exc()
        release_notes = 'failed to retrieve release-notes'

    if release_notes:
        max_pr_body_length = 65536 # also: max comment body length
        # If the size of the release-notes exceeds the max. body-length for PRs, split the notes
        # into MAX_PR_BODY_LENGTH-sized chunks and add subsequent chunks to the PR as comments.
        max_body_length_exceeded_remark = (
            '\n\nRelease notes were shortened since they exceeded the maximum length allowed for a '
            'pull request body. The remaining release notes will be added as comments to this PR.'
        )
        if max_pr_body_length < len(release_notes):
            step_size = max_pr_body_length - len(max_body_length_exceeded_remark)
            split_release_notes = [
                release_notes[start:start+step_size]
                for start in range(0, len(release_notes), step_size)
            ]
        else:
            split_release_notes = [release_notes]

        if not (additional_notes := split_release_notes[1:]):
            pr_body = split_release_notes[0]
        else:
            pr_body = split_release_notes[0] + max_body_length_exceeded_remark
    else:
        pr_body = ''
        additional_notes = []

    if pullrequest_body_suffix:
        pr_body += f'\n{pullrequest_body_suffix}'

    try:
        pull_request = ls_repo.create_pull(
            title=gu.PullRequestUtil.calculate_pr_title(
                reference=to_ref,
                from_version=from_version,
                to_version=to_version
            ),
            base=githubrepobranch.branch(),
            head=upgrade_branch_name,
            body=pr_body.strip(),
        )

        for release_note_part in additional_notes:
            pull_request.create_comment(body=release_note_part)
    except github3.exceptions.UnprocessableEntity as e:
        logger.info(f'Intercepted UnprocessableEntity exception. Listed errors: {e.errors}')
        raise

    if merge_policy is ucd.MergePolicy.MANUAL:
        return pull_request_util._pr_to_upgrade_pull_request(pull_request)

    logger.info(
        f"Merging upgrade-pr #{pull_request.number} ({merge_method=!s}) on branch "
        f"'{upgrade_branch_name}' into branch '{githubrepobranch.branch()}'."
    )

    def  _merge_pr(
        merge_method: ucd.MergeMethod,
        pull_request: github3.github.pulls.ShortPullRequest,
        attempts: int,
        delay: int = 2,
    ):
        if attempts > 0:
            try:
                if merge_method is ucd.MergeMethod.MERGE:
                    pull_request.merge(merge_method='merge')
                elif merge_method is ucd.MergeMethod.REBASE:
                    pull_request.merge(merge_method='rebase')
                elif merge_method is ucd.MergeMethod.SQUASH:
                    pull_request.merge(merge_method='squash')
                else:
                    raise NotImplementedError(f'{merge_method=}')
            except github3.exceptions.MethodNotAllowed as e:
                remaining_attempts = attempts-1
                logger.warning(
                    f'Encountered an exception when merging PR: {e}. Will wait {delay} seconds '
                    f'and try again {remaining_attempts} time(s).'
                )
                time.sleep(delay)
                _merge_pr(
                    merge_method=merge_method,
                    pull_request=pull_request,
                    attempts=remaining_attempts,
                )
        else:
            logger.warning(
                f'Unable to merge upgrade pull request #{pull_request.number} '
                f'({pull_request.html_url}).'
            )

    _merge_pr(merge_method=merge_method, pull_request=pull_request, attempts=3)

    try:
        ls_repo.ref(f'heads/{upgrade_branch_name}').delete()
    except github3.exceptions.NotFoundError:
        pass

    if after_merge_callback:
        subprocess.run(
            [os.path.join(repo_dir, after_merge_callback)],
            check=True,
            env=cmd_env
        )

    return pull_request_util._pr_to_upgrade_pull_request(pull_request)


def push_upgrade_commit(
    ls_repo: github3.repos.repo.Repository,
    commit_message: str,
    githubrepobranch: gu.GitHubRepoBranch,
    repo_dir: str,
) -> str:
    # mv diff into commit and push it
    helper = gitutil.GitHelper.from_githubrepobranch(
        githubrepobranch=githubrepobranch,
        repo_path=repo_dir,
    )
    commit = helper.index_to_commit(message=commit_message)
    logger.info(f'commit for upgrade-PR: {commit.hexsha=}')
    new_branch_name = ci.util.random_str(prefix='ci-', length=12)
    repo_branch = githubrepobranch.branch()
    head_sha = ls_repo.ref(f'heads/{repo_branch}').object.sha
    ls_repo.create_ref(f'refs/heads/{new_branch_name}', head_sha)

    try:
        helper.push(from_ref=commit.hexsha, to_ref=f'refs/heads/{new_branch_name}')
    except:
        logger.warning('an error occurred - removing now useless pr-branch')
        ls_repo.ref(f'heads/{new_branch_name}').delete()
        raise

    helper.repo.git.checkout('.')

    return new_branch_name


def create_release_notes(
    from_component: ocm.Component,
    from_github_cfg,
    from_repo_owner: str,
    from_repo_name: str,
    to_version: str,
    version_lookup,
    component_descriptor_lookup,
):
    from_version = from_component.version

    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            gitutil.GitHelper.clone_into(
                target_directory=temp_dir,
                github_cfg=from_github_cfg,
                github_repo_path=f'{from_repo_owner}/{from_repo_name}'
            )
            release_note_blocks = release_notes_fetch.fetch_release_notes(
                component=from_component,
                version_lookup=version_lookup,
                component_descriptor_lookup=component_descriptor_lookup,
                repo_path=temp_dir,
                current_version=to_version,
                previous_version=from_version,
            )
            if release_note_blocks:
                n = '\n'
                return f'**Release Notes**:\n{n.join(r.block_str for r in release_note_blocks)}'

    except:
        logger.warning('an error occurred during release notes processing (ignoring)')
        import traceback
        logger.warning(traceback.format_exc())



import base64
import logging
import random
import urllib3.exceptions

import kubernetes.client
import kubernetes.config

import ccc.delivery
import ccc.github
import ccc.secrets_server
import cfg_mgmt.model as cmm
import cfg_mgmt.reporting as cmr
import cfg_mgmt.util as cmu
import ci.log
import ci.util
import model
import model.concourse
import model.config_repo
import model.secret
import model.secrets_server

ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


def process_config_queue(
    cfg_dir: str,
    repo_url: str,
    github_repo_path: str,
    target_ref: str,
):
    '''
    Find first config queue entry that should be deleted and delete it.
    '''
    cfg_metadata = cmm.cfg_metadata_from_cfg_dir(cfg_dir=cfg_dir)
    cfg_factory = model.ConfigFactory.from_cfg_dir(
        cfg_dir=cfg_dir,
        disable_cfg_element_lookup=True,
    )
    github_cfg = ccc.github.github_cfg_for_repo_url(
        repo_url,
        cfg_factory=cfg_factory,
    )

    cfg_queue_entries = list(cmu.iter_cfg_queue_entries_to_be_deleted(cfg_metadata=cfg_metadata))
    random.shuffle(cfg_queue_entries) # do not block cfg-queue if entry cannot be processed

    for cfg_queue_entry in cfg_queue_entries:
        cfg_element = cfg_factory._cfg_element(
            cfg_type_name=cfg_queue_entry.target.type,
            cfg_name=cfg_queue_entry.target.name,
        )
        if cmu.process_cfg_queue_and_persist_in_repo(
            cfg_element=cfg_element,
            cfg_factory=cfg_factory,
            cfg_metadata=cfg_metadata,
            cfg_queue_entry=cfg_queue_entry,
            cfg_dir=cfg_dir,
            github_cfg=github_cfg,
            github_repo_path=github_repo_path,
            target_ref=target_ref,
        ):
            # stop after first successful deletion (avoid causing too much trouble at one time
            return
    logger.info('did not find a config queue entry to delete')


def rotate_secrets(
    cfg_dir: str,
    target_ref: str,
    repo_url: str,
    github_repo_path: str,
):
    cfg_metadata = cmm.cfg_metadata_from_cfg_dir(cfg_dir=cfg_dir)
    cfg_factory = model.ConfigFactory.from_cfg_dir(
        cfg_dir=cfg_dir,
        disable_cfg_element_lookup=True,
    )
    github_cfg = ccc.github.github_cfg_for_repo_url(
        repo_url,
        cfg_factory=cfg_factory,
    )

    cfg_elements_requiring_rotation = list(cmr.iter_cfg_elements_requiring_rotation(
        cmu.iter_cfg_elements(cfg_factory=cfg_factory),
        cfg_metadata=cfg_metadata,
        rotation_method=cmm.RotationMethod.AUTOMATED,
    ))

    # do not block queue if element cannot be processed
    random.shuffle(cfg_elements_requiring_rotation)

    for cfg_element in cfg_elements_requiring_rotation:
        logger.info(
            f"Rotating config-element '{cfg_element.name()}' of type '{cfg_element._type_name}'"
        )
        if cmu.rotate_config_element_and_persist_in_cfg_repo(
            cfg_element=cfg_element,
            cfg_factory=cfg_factory,
            cfg_metadata=cfg_metadata,
            cfg_dir=cfg_dir,
            github_cfg=github_cfg,
            github_repo_path=github_repo_path,
            target_ref=target_ref,
        ):
            # stop after first successful rotation (avoid causing too much trouble at one time)
            break


def _put_secret(
        core_api: kubernetes.client.CoreV1Api,
        name: str,
        data: dict = None,
        namespace: str='default',
        raw_data: dict = None,
):
    '''creates or updates (replaces) the specified secret.
    the secret's contents are expected in a dictionary containing only scalar values.
    In particular, each value is converted into a str; the result returned from
    to-str conversion is encoded as a utf-8 byte array. Thus such a conversion must
    not have done before.
    '''
    if not bool(data) ^ bool(raw_data):
        raise ValueError('Exactly one data or raw data has to be set')

    metadata = kubernetes.client.V1ObjectMeta(
        name=ci.util.not_empty(name),
        namespace=ci.util.not_empty(namespace),
    )

    if data:
        raw_data = {
            k: base64.b64encode(str(v).encode('utf-8')).decode('utf-8')
            for k,v in data.items()
        }

    secret = kubernetes.client.V1Secret(metadata=metadata, data=raw_data)

    # find out whether we have to replace or to create
    try:
        core_api.read_namespaced_secret(name=name, namespace=namespace)
        secret_exists = True
    except kubernetes.client.ApiException as ae:
        # only 404 is expected
        if not ae.status == 404:
            raise ae
        secret_exists = False

    if secret_exists:
        core_api.replace_namespaced_secret(name=name, namespace=namespace, body=secret)
    else:
        core_api.create_namespaced_secret(namespace=namespace, body=secret)


def _put_k8s_secret(
    cfg_factory: model.ConfigFactory,
    target: model.config_repo.KubernetesSecretTarget,
    cfg_sets: list[model.ConfigurationSet],
):
    kubernetes_config = cfg_factory.kubernetes(target.kubernetes_config_name)
    target_kubeconfig = kubernetes_config.kubeconfig()

    secret_name = target.secret_name
    secret_namespace = target.secret_namespace
    secret_key = target.secret_key

    serialiser = model.ConfigSetSerialiser(cfg_sets=cfg_sets, cfg_factory=cfg_factory)
    serialised = serialiser.serialise().encode('utf-8')
    secret_data = base64.b64encode(serialised).decode('utf-8')

    api_client = kubernetes.config.new_client_from_config_dict(target_kubeconfig)
    core_api = kubernetes.client.CoreV1Api(api_client)

    logger.info(
        f"deploying config into k8s-secret '{secret_name}' in namespace "
        f"'{secret_namespace}' on cluster '{api_client.configuration.host}'"
    )

    try:
        _put_secret(
            core_api=core_api,
            name=secret_name,
            raw_data={secret_key: secret_data},
            namespace=secret_namespace,
        )
    except urllib3.exceptions.MaxRetryError as e:
        logger.error(
            f'cannot update k8s-secret on cluster {api_client.configuration.host} '
            f'- maybe it is hibernated? Will continue with remaining secret replication; {e}'
        )


def _put_secrets_server_secret(
    cfg_factory: model.ConfigFactory,
    target: model.config_repo.SecretsServerTarget,
    cfg_sets: list[model.ConfigurationSet],
):
    kubernetes_config = cfg_factory.kubernetes(target.kubernetes_config_name)
    target_kubeconfig = kubernetes_config.kubeconfig()

    secret_config: model.secret.Secret = cfg_factory.secret(target.secret_config)
    secrets_server_config: model.secrets_server.SecretsServerConfig = \
        cfg_factory.secrets_server(target.secrets_server_config)
    logger.info(
        f'secret cfg: {secret_config.name()} '
        f'and key: {secret_config.key().decode("utf-8")[:3]}...'
    )

    team_cfg = cfg_factory.concourse_team_cfg(target.team_config)
    concourse_team_name = team_cfg.team_name()

    secret_name = model.concourse.secret_name_from_team(
        team_name=concourse_team_name,
        key_generation=secret_config.generation(),
    )
    secret_namespace = secrets_server_config.namespace()
    secret_key = model.concourse.secret_cfg_name_for_team(concourse_team_name)

    serialiser = model.ConfigSetSerialiser(cfg_sets=cfg_sets, cfg_factory=cfg_factory)
    encrypted_cipher_data = ccc.secrets_server.encrypt_data(
        key=secret_config.key(),
        cipher_algorithm=secret_config.cipher_algorithm(),
        serialized_secret_data=serialiser.serialise().encode('utf-8')
    )
    secret_data = base64.b64encode(encrypted_cipher_data).decode('utf-8')

    api_client = kubernetes.config.new_client_from_config_dict(target_kubeconfig)
    core_api = kubernetes.client.CoreV1Api(api_client)

    logger.info(
        f"deploying encrypted secret '{secret_name}' in namespace '{secret_namespace}' "
        f"on cluster '{api_client.configuration.host}'"
    )
    _put_secret(
        core_api=core_api,
        name=secret_name,
        raw_data={secret_key: secret_data},
        namespace=secret_namespace,
    )


def replicate_secrets(
    cfg_factory: model.ConfigFactory,
    replication_target_config: model.config_repo.ReplicationTargetConfig,
):
    for mapping in replication_target_config.replication_mappings:
        cfg_set = cfg_factory.cfg_set(mapping.cfg_set)
        # force cfg_set serialiser to include referenced cfg_sets
        cfg_sets = list(cfg_set._cfg_elements('cfg_set')) + [cfg_set]
        if isinstance(target := mapping.target, model.config_repo.KubernetesSecretTarget):
            _put_k8s_secret(
                cfg_factory=cfg_factory,
                target=target,
                cfg_sets=cfg_sets,
            )
        elif isinstance(target, model.config_repo.SecretsServerTarget):
            _put_secrets_server_secret(
                cfg_factory=cfg_factory,
                target=target,
                cfg_sets=cfg_sets,
            )
        else:
            raise NotImplementedError(type(mapping))



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import enum
import os
import string
import shlex

import ci.util
import model.concourse

from concourse.model.base import (
    AttributeSpec,
    ModelBase,
    ModelValidationError,
    ScriptType,
    normalise_to_dict,
)


def from_instance(attr_name, value_doc: str=None):
    '''
    returns an attribute from a concourse.model.ModelBase instance

    Used as default value for PipelineStep's execute attribute to support dynamic default values.
    '''
    def get_attr(self: ModelBase):
        return getattr(self, attr_name)

    if value_doc is not None:
        get_attr.__name__ = value_doc
    else:
        get_attr.__name__ = f'{attr_name}'

    return get_attr


class PrivilegeMode(enum.Enum):
    PRIVILEGED = 'privileged'
    UNPRIVILEGED = 'unprivileged'


def attrs(pipeline_step):
    return (
        AttributeSpec.optional(
            name='depends',
            default=set(),
            doc='step names this step declares a dependency towards',
            type=set,
        ),
        AttributeSpec.optional(
            name='trait_depends',
            default=set(),
            doc=(
                'names of traits this step declares a dependency towards. This will result in a '
                'dependency towards all steps defined by these traits'
            ),
            type=set,
        ),
        AttributeSpec.optional(
            name='execute',
            default=from_instance(attr_name='name', value_doc='step name'),
            doc='''
            The executable (with optional additional arguments) to run. The executable path
            is calculated relative to `<main_repo>/.ci`.

            Has two forms:

            - scalar value (str in most cases) --> no shell-escaping is done
            - list of scalar values -> used verbatim as ARGV

            ''',
        ),
        AttributeSpec.optional(
            name='escape_argv',
            default=True,
            doc='''\
                defines whether or not ARGV (execute attr) should be escaped
                setting to `False` will allow for e.g. environment-variable expansion
            ''',
        ),
        AttributeSpec.optional(
            name='notifications_cfg',
            default='default',
            doc='''
            Configures build notification policies (see
            :ref:`notifications trait <trait-notifications>`)
            ''',
            type=str,
        ),
        AttributeSpec.optional(
            name='image',
            default=None,
            doc='''
            the container image reference to use for the executing container.
            If not set, the default image will be used.
            ''',
        ),
        AttributeSpec.optional(
            name='registry',
            default=None,
            doc='''
            The container image registry cfg_name. Required when retrieving container images
            from a non-default image registry that requires authentication.
            ''',
        ),
        AttributeSpec.optional(
            name='inputs',
            default={},
            doc='''
            a mapping of inputs produced by other build steps:
            { input_name: output_name }
            `input_name` is converted to UPPER_CASE and exposed to the step as an environment
            variable containing the relative path to the output.
            ''',
            type=dict,
        ),
        AttributeSpec.optional(
            name='output_dir',
            default=None,
            doc='''
            exposes a writable directory to the job. The directory is specified via environment
            variable named as the given value + _PATH (converted to UPPER-case and snake_case).
            Any files placed into this directory are passed to subsequent steps declaring the output
            as input. The unchanged value configured is used as input name.
            e.g.: `output_dir: out` results in env var `OUT_PATH`.
            ''',
        ),
        AttributeSpec.optional(
            name='publish_to',
            default={},
            doc='''
            has two forms:

            * list of logical repository names to which commits created by this step should be
              published.
            * a dictionary: <name: options>

            The second form currently accepts exactly one argument: `force_push` (bool) and is used
            to specify that a force-push should be done.

            The step executable must only commit the changes in the repository's worktree without
            pushing them.

            Example:

            .. code-block:: yaml

                steps:
                    foo:
                        publish_to:
                            my_repo:
                                force_push: true
            ''',
            type=list,
        ),
        AttributeSpec.optional(
            name='vars',
            default={},
            doc='''
            pairs of {env_var_name: <python expression>}
            the specified python expressions are evaluated during pipeline replication.
            An instance of the current pipeline_model is accessible through the
            `pipeline_descriptor` symbol.
            The evaluation result is exposed to this build step via the specified environment
            variable.
            ''',
        ),
        AttributeSpec.optional(
            name='cache_paths',
            default=[],
            doc='''
            .. warning::
                EXPERIMENTAL - this attr might be removed or changed

            a list of paths (relative to initial PWD) that should be cached.
            ''',
        ),
        AttributeSpec.optional(
            name='privilege_mode',
            default=PrivilegeMode.UNPRIVILEGED,
            type=PrivilegeMode,
            doc='''
            privilege mode for step. Use carefully when running potentially untrusted code.
            ''',
        ),
        AttributeSpec.optional(
            name='timeout',
            default=None,
            doc='''
            go-style time interval (e.g.: '1h30m') after which the step will be interrupted and fail.
            ''',
        ),
        AttributeSpec.optional(
            name='retries',
            default=None,
            doc='''
            positive integer specifying the maximum amount of failures until the step is
            counted as failed
            ''',
        ),
        AttributeSpec.optional(
            name='on_abort',
            default=None,
            doc='''
            The executable (with optional additional arguments) to run in case the job is aborted.
            The executable path is calculated relative to `<main_repo>/.ci`.

            Just like `execute` this has two forms:

            - scalar value (str in most cases) --> no shell-escaping is done
            - list of scalar values -> used verbatim as ARGV

            The resulting `on_abort` step will use the same container image reference as the job
            that defines it. Also, it will have the same in- and outputs available and access to
            the same env-vars.

            .. note::
                `on_abort`-steps themselves cannot be aborted. Beware.

            ''',
        ),
    )


class StepNotificationPolicy(enum.Enum):
    NO_NOTIFICATION = enum.auto()
    ALWAYS = enum.auto()


class PullRequestNotificationPolicy(enum.Enum):
    NO_NOTIFICATION = enum.auto()
    ALWAYS = enum.auto()


class TaskHook(enum.Enum):
    NONE = enum.auto()
    ON_ABORT = enum.auto()


class PipelineStep(ModelBase):
    def __init__(
        self,
        name: str,
        is_synthetic: bool,
        script_type: ScriptType,
        notification_policy: StepNotificationPolicy = StepNotificationPolicy.ALWAYS,
        pull_request_notification_policy: PullRequestNotificationPolicy = (
            PullRequestNotificationPolicy.ALWAYS
        ),
        extra_args=None,
        injecting_trait_name=None,
        worker_node_tags: tuple[str]=(),
        platform: model.concourse.Platform=None,
        *args,
        **kwargs
    ):
        '''
        A single pipeline step.

        name: step's name (displayed to end-users + used e.g. for retrieving logs)
        is_synthetic: if True, step was injected by a trait; otherwise by user from
                      pipeline-definition (by end-user)
        script_type: influences which executable is called for executing body
        notification_policy: see enum-values
        extra_args: passed to step-specific template-processing during render-time
        injecting_trait_name: name of the trait that injected this ("synthetic") step
        worker_node_tags: if set, step will be restricted to run on worker-nodes bearing all of
                          the given tags
        '''
        self.name = ci.util.not_empty(name)
        self.is_synthetic = is_synthetic
        self._script_type = script_type
        self._outputs_dict = {}
        self._inputs_dict = {}
        self._publish_to_dict = {}
        self._notification_policy = notification_policy
        self._pull_request_notification_policy = pull_request_notification_policy
        self._notifications_cfg = None
        self._injecting_trait_name = injecting_trait_name
        self._extra_args = extra_args
        self._worker_node_tags = tuple(worker_node_tags)
        self._platform = platform
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return attrs(cls)

    @classmethod
    def _defaults_dict(cls):
        return AttributeSpec.defaults_dict(attrs(cls))

    @classmethod
    def _optional_attributes(cls):
        return set(AttributeSpec.optional_attr_names(attrs(cls)))

    @classmethod
    def _required_attributes(cls):
        return set(AttributeSpec.required_attr_names(attrs(cls)))

    def custom_init(self, raw_dict: dict):
        if not isinstance(raw_dict, dict):
            raise ValueError(f'expected a dict, but received: {type(raw_dict)} ({raw_dict})')

        raw_dict['depends'] = set(raw_dict['depends'])
        if (name := raw_dict.get('output_dir', None)):
            self.add_output(
                name=name,
                variable_name=f'{name}_path',
            )

        # add hard-coded output "on_error" (allows build steps to pass custom
        # notification cfg for build errors)
        self.add_output(name='on_error_dir', variable_name='on_error_dir')

        for variable_name, name in raw_dict.get('inputs').items():
            self.add_input(name=name, variable_name=variable_name)

        self._publish_to_dict = normalise_to_dict(raw_dict['publish_to'])

    def script_type(self) -> ScriptType:
        '''
        returns the step's "script type"

        The script type specifies the execution environment in which the script payload is run
        (script payloads are hard-coded in pipeline templates).
        '''
        return self._script_type

    def notifications_cfg_name(self):
        return self.raw['notifications_cfg']

    def notifications_cfg(self):
        # injected by notifications trait
        return self._notifications_cfg

    def image(self):
        return self.raw['image']

    @property
    def worker_node_tags(self):
        return self._worker_node_tags

    @property
    def platform(self) -> model.concourse.Platform | None:
        return self._platform

    @property
    def escape_argv(self):
        return self.raw.get('escape_argv', True)

    def _execute(self):
        # by default, run an executable named as the step
        execute_value = self.raw.get('execute', self.name)
        if callable(execute_value):
            return execute_value(self)
        else:
            return execute_value

    def _argv(self, hook: TaskHook = TaskHook.NONE):
        if hook is TaskHook.NONE:
            execute = self._execute()
        elif hook is TaskHook.ON_ABORT:
            execute = self._on_abort()
        else:
            raise NotImplementedError(hook)

        if execute is None:
            # This can only happen if the template erroneously calls for the task-hook to be
            # rendered. Check the template for errors.
            raise RuntimeError(f'Nothing to execute configured for {hook}')

        if not isinstance(execute, list):
            return [str(execute)]

        if self.escape_argv:
            return [shlex.quote(str(e)) for e in execute]
        else:
            return execute

    def executable(self, prefix='', hook: TaskHook = TaskHook.NONE):
        argv = self._argv(hook)
        if isinstance(prefix, str):
            prefix = [prefix]
        return os.path.join(*prefix, argv[0]).rstrip()

    def execute(self, prefix='', hook: TaskHook = TaskHook.NONE):
        argv = self._argv(hook)
        argv[0] = self.executable(prefix=prefix, hook=hook)
        return ' '.join(argv).rstrip()

    def registry(self):
        return self.raw.get('registry', None)

    def output_dir(self):
        return self.raw['output_dir']

    def output(self, name):
        outputs = self.outputs()
        if name not in outputs:
            raise ValueError(f'{name} not found in {list(outputs.keys())}')
        return self.outputs()[name]

    def outputs(self):
        return self._outputs_dict

    def add_output(self, name, variable_name):
        if variable_name in self._outputs_dict:
            raise ValueError(f'output already exists: {variable_name}')
        self._outputs_dict[variable_name] = name

    def inputs(self):
        return self._inputs_dict

    def input(self, name):
        return self.inputs()[name]

    def add_input(self, name, variable_name):
        ci.util.not_none(name)
        ci.util.not_none(variable_name)

        if variable_name in self._inputs_dict and name != self._inputs_dict[variable_name]:
            raise ValueError(f'conflict: {variable_name} was tried to be reset to different value')
        self._inputs_dict[variable_name] = name

    def remove_input(self, name):
        ci.util.not_none(name)

        if not name in self._inputs_dict:
            raise ValueError(f'input does not exist: {name}')
        self._inputs_dict.pop(name)

    def cache_paths(self):
        return self.raw['cache_paths']

    def variables(self):
        return self.raw.get('vars')

    def publish_repository_names(self):
        return self._publish_to_dict.keys()

    def publish_repository_dict(self):
        return self._publish_to_dict

    def _add_dependency(self, step: 'PipelineStep'):
        if step.name == self.name:
            return # ignore dependencies towards self
        self.raw['depends'].add(step.name)

    def _remove_dependency(self, step: 'PipelineStep'):
        self.raw['depends'].remove(step.name)

    def depends(self):
        return set(self.raw['depends'])

    def trait_depends(self):
        return set(self.raw['trait_depends'])

    def _dependencies(self) -> set[str]:
        return self.raw['depends']

    def injecting_trait_name(self):
        return self._injecting_trait_name

    def timeout(self):
        return self.raw['timeout']

    def set_timeout(self, duration_string: str):
        ci.util.not_empty(duration_string)
        self.raw['timeout'] = duration_string

    def privilege_mode(self) -> PrivilegeMode:
        return PrivilegeMode(self.raw['privilege_mode'])

    def retries(self):
        return self.raw['retries']

    def notification_policy(self):
        return self._notification_policy

    def _on_abort(self):
        return self.raw.get('on_abort', None)

    def pull_request_notification_policy(self):
        return self._pull_request_notification_policy

    def validate(self):
        super().validate()
        if self.image():
            image_reference = self.image()
            # image must be a valid oci image reference
            allowed_characters = string.ascii_letters + string.digits + '.-_/:@'
            if any(map(lambda c: c not in allowed_characters, image_reference)):
                raise ModelValidationError(
                    'forbidden character in image reference: ' + str(image_reference)
                )
            if ':' not in image_reference:
                raise ModelValidationError(
                    'image reference must contain colon charater:' + str(image_reference)
                )

    def __str__(self):
        descr = 'PipelineStep {n} - depends: {d}, inputs: {i}, outputs: {o}'.format(
            n=self.name,
            d=self.depends(),
            i=self.inputs(),
            o=self.outputs(),
        )
        return descr



import collections.abc
import dataclasses
import io
import itertools
import json
import logging
import os
import shutil
import tarfile
import tempfile

import requests
import yaml

import cachetools
import dacite
import ocm
import gci.oci

import ci.util
import cnudie.util
import oci.client as oc
import oci.model as om


logger = logging.getLogger(__name__)

ComponentName = str | tuple[str, str] | ocm.Component | ocm.ComponentIdentity

ComponentDescriptorLookupById = collections.abc.Callable[
    [ocm.ComponentIdentity, ocm.OcmRepository],
    ocm.ComponentDescriptor
]

VersionLookupByComponent = collections.abc.Callable[
    [ComponentName, ocm.OcmRepository],
    collections.abc.Sequence[str]
]


@dataclasses.dataclass(frozen=True)
class OcmRepositoryMappingEntry:
    repository: str
    prefix: str | None = None


OcmRepositoryCfg = str | collections.abc.Iterable[str]


OcmRepositoryLookup = collections.abc.Callable[
    [ComponentName],
    collections.abc.Generator[ocm.OciOcmRepository | str, None, None],
]


def _iter_ocm_repositories(
    component: str | ocm.ComponentIdentity | ocm.Component,
    repository_cfg: OcmRepositoryCfg,
    /,
):
    if isinstance(component, ocm.ComponentIdentity):
        component = component.name
    elif isinstance(component, ocm.Component):
        component = component.name

    if isinstance(repository_cfg, ocm.OciOcmRepository):
        yield repository_cfg
        return

    if isinstance(repository_cfg, str):
        yield repository_cfg
        return

    # recurse into elements in case repository_cfg is iterable
    if hasattr(repository_cfg, '__iter__'):
        for cfg in repository_cfg:
            yield _iter_ocm_repositories(component, cfg)
        return


def iter_ocm_repositories(
    component: str | ocm.ComponentIdentity | ocm.Component,
    /,
    *repository_cfgs: OcmRepositoryCfg,
) -> collections.abc.Generator[ocm.OciOcmRepository | str, None, None]:
    for repository_cfg in repository_cfgs:
        if not repository_cfg:
            # convenience: ignore, e.g. None
            # handy for passing-in multiple values, from which some may be none
            continue
        if callable(repository_cfg):
            for cfg in repository_cfg(component):
                if not cfg:
                    continue
                yield cfg
            return

        for cfg in _iter_ocm_repositories(component, repository_cfg):
            if hasattr(cfg, '__iter__'):
                yield from cfg
            else:
                yield cfg


def ocm_repository_lookup(*repository_cfgs: OcmRepositoryCfg):
    def lookup(
        component: str | ocm.ComponentIdentity | ocm.Component,
    ):
        return iter_ocm_repositories(component, repository_cfgs)

    return lookup


class WriteBack:
    '''
    Wraps a writeback function which can be used to fill the lookup in case the required
    component descriptor was not found.
    '''
    def __init__(
        self,
        writeback: collections.abc.Callable[[ocm.ComponentIdentity, ocm.ComponentDescriptor], None],
    ):
        self.writeback = writeback

    def __call__(
        self,
        component_id: ocm.ComponentIdentity,
        component_descriptor: ocm.ComponentDescriptor,
    ):
        self.writeback(component_id, component_descriptor)


def in_memory_cache_component_descriptor_lookup(
    cache_ctor: cachetools.Cache=cachetools.LRUCache,
    ocm_repository_lookup: OcmRepositoryLookup=None,
    **cache_kwargs,
) -> ComponentDescriptorLookupById:
    '''
    Used to lookup referenced component descriptors in the in-memory cache.
    In case of a cache miss, the required component descriptor can be added
    to the cache by using the writeback function.

    @param cache_ctor:
        specification of the cache implementation
    @param ocm_repository_lookup:
        lookup for OCM repositories
    @param cache_kwargs:
        further args used for cache initialization, maxsize is defaulted to 2048
    '''
    cache_kwargs['maxsize'] = cache_kwargs.get('maxsize', 2048)
    cache = cache_ctor(**cache_kwargs)

    def writeback(
        component_id: ocm.ComponentIdentity,
        component_descriptor: ocm.ComponentDescriptor,
    ):
        if (ocm_repo := component_descriptor.component.current_ocm_repo):
            cache.__setitem__((component_id, ocm_repo), component_descriptor)
        else:
            raise ValueError(ocm_repo)

    _writeback = WriteBack(writeback)

    def lookup(
        component_id: ocm.ComponentIdentity,
        ctx_repo: ocm.OcmRepository=None,
        ocm_repository_lookup=ocm_repository_lookup,
    ):
        if ctx_repo:
            ocm_repos = (ctx_repo,)

        else:
            ocm_repos = iter_ocm_repositories(
                component_id,
                ocm_repository_lookup,
            )

        for ocm_repo in ocm_repos:
            if isinstance(ocm_repo, str):
                ocm_repo = ocm.OciOcmRepository(
                    type=ocm.AccessType.OCI_REGISTRY,
                    baseUrl=ocm_repo,
                )
            try:
                if (component_descriptor := cache.get((component_id, ocm_repo))):
                    return component_descriptor
            except KeyError:
                pass

        # component descriptor not found in lookup
        return _writeback

    return lookup


def file_system_cache_component_descriptor_lookup(
    ocm_repository_lookup: OcmRepositoryLookup=None,
    cache_dir: str=None,
) -> ComponentDescriptorLookupById:
    '''
    Used to lookup referenced component descriptors in the file-system cache.
    In case of a cache miss, the required component descriptor can be added
    to the cache by using the writeback function. If cache_dir is not specified,
    it is tried to retrieve it from configuration (see `ctx`).

    @param ocm_repository_lookup:
        lookup for OCM repositories
    @param cache_dir:
        directory used for caching. Must exist, otherwise a ValueError is raised
    '''
    if not cache_dir:
        raise ValueError(cache_dir)

    def writeback(
        component_id: ocm.ComponentIdentity,
        component_descriptor: ocm.ComponentDescriptor,
    ):
        if not (ocm_repo := component_descriptor.component.current_ocm_repo):
            raise ValueError(ocm_repo)

        try:
            f = tempfile.NamedTemporaryFile(mode='w', delete=False)
            # write to tempfile, followed by a mv to avoid collisions through concurrent
            # processes or threads (assuming mv is an atomic operation)
            yaml.dump(
                data=dataclasses.asdict(component_descriptor),
                Dumper=ocm.EnumValueYamlDumper,
                stream=f.file,
            )
            f.close() # need to close filehandle for NT

            descriptor_path = os.path.join(
                cache_dir,
                ocm_repo.oci_ref.replace('/', '-'),
                f'{component_id.name}-{component_id.version}',
            )
            if not os.path.isfile(descriptor_path):
                base_dir = os.path.dirname(descriptor_path)
                os.makedirs(name=base_dir, exist_ok=True)
            shutil.move(f.name, descriptor_path)
        except:
            os.unlink(f.name)
            raise

    _writeback = WriteBack(writeback)

    def lookup(
        component_id: cnudie.util.ComponentId,
        ctx_repo: ocm.OcmRepository|str=None,
        ocm_repository_lookup: OcmRepositoryLookup=ocm_repository_lookup,
    ):
        if ctx_repo:
            ocm_repos = (ctx_repo, )
        else:
            ocm_repos = iter_ocm_repositories(
                component_id,
                ocm_repository_lookup,
            )

        for ocm_repo in ocm_repos:
            if not ocm_repo:
                raise ValueError(ocm_repo)

            if isinstance(ocm_repo, str):
                ocm_repo = ocm.OciOcmRepository(
                    type=ocm.AccessType.OCI_REGISTRY,
                    baseUrl=ocm_repo,
                )

            if not isinstance(ocm_repo, ocm.OciOcmRepository):
                raise NotImplementedError(ocm_repo)

            component_id = cnudie.util.to_component_id(component_id)

            descriptor_path = os.path.join(
                cache_dir,
                ocm_repo.oci_ref.replace('/', '-'),
                f'{component_id.name}-{component_id.version}',
            )
            if os.path.isfile(descriptor_path):
                return ocm.ComponentDescriptor.from_dict(
                    ci.util.parse_yaml_file(descriptor_path)
                )

        # component descriptor not found in lookup
        return _writeback

    return lookup


def delivery_service_component_descriptor_lookup(
    ocm_repository_lookup: OcmRepositoryLookup=None,
    delivery_client=None,
    default_absent_ok: bool=True,
    default_ignore_errors: tuple[Exception]=(
        requests.exceptions.ConnectionError,
        requests.exceptions.ReadTimeout,
    ),
    fallback_to_service_mapping: bool=True,
) -> ComponentDescriptorLookupById:
    '''
    Used to lookup referenced component descriptors in the delivery-service.

    @param ocm_repository_lookup:
        lookup for OCM repositories
    @param delivery_client:
        client to establish the connection to the delivery-service. If the client cannot be created,
        a ValueError is raised
    @param default_absent_ok:
        sets the default behaviour in case of absent component descriptors for the returned lookup
        function
    @param default_ignore_errors:
        collection of exceptions which should be ignored by default. In case of such an exception,
        no component descriptor will be returned, so that a subsequent lookup can retry retrieving
        it
    @param fallback_to_service_mapping:
        if set, it is tried to retrieve the requested component descriptor using the OCM repository
        mapping of the  delivery-service, in case it could not be retrieved using
        `ocm_repository_lookup`
    '''
    if not delivery_client:
        import ccc.delivery
        delivery_client = ccc.delivery.default_client_if_available()
    if not delivery_client:
        raise ValueError(delivery_client)

    def lookup(
        component_id: ocm.ComponentIdentity,
        ctx_repo: ocm.OcmRepository=None,
        ocm_repository_lookup: OcmRepositoryLookup=ocm_repository_lookup,
        absent_ok: bool=default_absent_ok,
        ignore_errors: tuple[Exception]=default_ignore_errors,
    ):
        component_id = cnudie.util.to_component_id(component_id)
        if ctx_repo:
            ocm_repos = (ctx_repo, )
        else:
            ocm_repos = iter_ocm_repositories(
                component_id,
                ocm_repository_lookup,
            )

        # if component descriptor is not found in `ocm_repos`, fallback to default ocm repo mapping
        # defined in delivery service (i.e. specify no ocm repository)
        if fallback_to_service_mapping:
            ocm_repos = itertools.chain(ocm_repos, (None,))

        for ocm_repo in ocm_repos:
            if isinstance(ocm_repo, str):
                ocm_repo = ocm.OciOcmRepository(
                    type=ocm.AccessType.OCI_REGISTRY,
                    baseUrl=ocm_repo,
                )

            if ocm_repo and not isinstance(ocm_repo, ocm.OciOcmRepository):
                raise NotImplementedError(ocm_repo)

            try:
                component_descriptor = delivery_client.component_descriptor(
                    name=component_id.name,
                    version=component_id.version,
                    ocm_repo_url=ocm_repo.oci_ref if ocm_repo else None,
                )

                if component_descriptor:
                    return component_descriptor
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404:
                    continue
                elif e.response.status_code >= 500:
                    # in case delivery-service is not reachable, fallback to next lookup (if any)
                    return None
                raise
            except ignore_errors:
                # already return here to not use unintended "fallback" ocm repositories
                return None

        # component descriptor not found in lookup
        if absent_ok:
            return None
        raise om.OciImageNotFoundException

    return lookup


def raw_component_descriptor_from_oci(
    component_id: ocm.ComponentIdentity,
    ocm_repos: collections.abc.Iterable[ocm.OciOcmRepository | str],
    oci_client: oc.Client=None,
    absent_ok: bool=False,
) -> bytes:
    if not oci_client:
        import ccc.oci
        if not (oci_client := ccc.oci.oci_client()):
            raise ValueError('oci_client must not be empty')

    for ocm_repo in ocm_repos:
        if isinstance(ocm_repo, str):
            ocm_repo = ocm.OciOcmRepository(
                type=ocm.OciAccess,
                baseUrl=ocm_repo,
            )

        if not isinstance(ocm_repo, ocm.OciOcmRepository):
            raise NotImplementedError(ocm_repo)

        target_ref = ci.util.urljoin(
            ocm_repo.oci_ref,
            'component-descriptors',
            f'{component_id.name.lower()}:{component_id.version}', # oci-spec allows only lowercase
        )

        manifest = oci_client.manifest(
            image_reference=target_ref,
            absent_ok=True,
        )

        if manifest:
            break
    else:
        manifest = None

    if not manifest and absent_ok:
        return None
    elif not manifest:
        raise om.OciImageNotFoundException

    try:
        cfg_dict = json.loads(
            oci_client.blob(
                image_reference=target_ref,
                digest=manifest.config.digest,
            ).text
        )
        cfg = dacite.from_dict(
            data_class=gci.oci.ComponentDescriptorOciCfg,
            data=cfg_dict,
        )
        layer_digest = cfg.componentDescriptorLayer.digest
        layer_mimetype = cfg.componentDescriptorLayer.mediaType
    except Exception as e:
        logger.warning(
            f'Failed to parse or retrieve component-descriptor-cfg: {e=}. '
            'falling back to single-layer'
        )

        # by contract, there must be exactly one layer (tar w/ component-descriptor)
        if not (layers_count := len(manifest.layers) == 1):
            logger.warning(f'XXX unexpected amount of {layers_count=}')

        layer_digest = manifest.layers[0].digest
        layer_mimetype = manifest.layers[0].mediaType

    if not layer_mimetype in gci.oci.component_descriptor_mimetypes:
        logger.warning(f'{target_ref=} {layer_mimetype=} was unexpected')
        # XXX: check for non-tar-variant

    return oci_client.blob(
        image_reference=target_ref,
        digest=layer_digest,
        stream=False, # manifests are typically small - do not bother w/ streaming
    ).content


def oci_component_descriptor_lookup(
    ocm_repository_lookup: OcmRepositoryLookup=None,
    oci_client: oc.Client | collections.abc.Callable[[], oc.Client]=None,
    default_absent_ok=True,
) -> ComponentDescriptorLookupById:
    '''
    Used to lookup referenced component descriptors in the oci-registry.

    @param ocm_repository_lookup:
        lookup for OCM repositories
    @param oci_client:
        client to establish the connection to the oci-registry. If the client cannot be created, a
        ValueError is raised
    @param default_absent_ok:
        sets the default behaviour in case of absent component descriptors for the returned lookup
        function
    '''
    if not oci_client:
        import ccc.oci
        oci_client = ccc.oci.oci_client()
    if not oci_client:
        raise ValueError(oci_client)

    def lookup(
        component_id: ocm.ComponentIdentity,
        ctx_repo: ocm.OcmRepository=None,
        ocm_repository_lookup: OcmRepositoryLookup=ocm_repository_lookup,
        absent_ok=default_absent_ok,
    ):
        component_id = cnudie.util.to_component_id(component_id)

        if isinstance(oci_client, collections.abc.Callable):
            local_oci_client = oci_client()
        else:
            local_oci_client = oci_client

        if ctx_repo:
            ocm_repos = (ctx_repo,)
        else:
            if not ocm_repository_lookup:
                raise ValueError('either ctx_repo, or ocm_repository_lookup must be passed')

            ocm_repos = iter_ocm_repositories(
                component_id,
                ocm_repository_lookup,
            )

        raw = raw_component_descriptor_from_oci(
            component_id=component_id,
            ocm_repos=ocm_repos,
            oci_client=local_oci_client,
            absent_ok=absent_ok,
        )
        if not raw and absent_ok:
            return
        elif not raw and not absent_ok:
            raise om.OciImageNotFoundException(component_id)

        # wrap in fobj
        blob_fobj = io.BytesIO(raw)
        try:
            component_descriptor = gci.oci.component_descriptor_from_tarfileobj(
                fileobj=blob_fobj,
            )
        except tarfile.ReadError as tre:
            tre.add_note(f'{component_id=}')
            raise tre

        return component_descriptor

    return lookup


def error_code_indicating_not_found(image_reference: str | om.OciImageReference) -> int:
    '''
    Since some oci registries don't comply with the open containers spec
    (https://github.com/opencontainers/distribution-spec/blob/main/spec.md#endpoints)
    regarding the error codes for failure upon listing tags, this function tries to
    guess the oci registry type based on the supplied `image_reference` and returns
    the expected error code.
    '''
    oci_registry_type = om.OciRegistryType.from_image_ref(image_reference=image_reference)

    if oci_registry_type == om.OciRegistryType.DOCKERHUB:
        return 401
    if oci_registry_type == om.OciRegistryType.GHCR:
        return 403

    return 404


def version_lookup(
    ocm_repository_lookup: OcmRepositoryLookup=None,
    oci_client: oc.Client=None,
    default_absent_ok: bool=True,
) -> VersionLookupByComponent:
    if not oci_client:
        import ccc.oci
        oci_client = ccc.oci.oci_client()
    if not oci_client:
        raise ValueError(oci_client)

    def lookup(
        component_id: ComponentName,
        ctx_repo: ocm.OcmRepository=None,
        ocm_repository_lookup: OcmRepositoryLookup=ocm_repository_lookup,
        absent_ok: bool=default_absent_ok,
    ):
        component_name = cnudie.util.to_component_name(component_id)
        if ctx_repo:
            ocm_repos = (ctx_repo, )
        else:
            ocm_repos = iter_ocm_repositories(
                component_name,
                ocm_repository_lookup,
            )

        versions = set()
        for ocm_repo in ocm_repos:
            if isinstance(ocm_repo, str):
                ocm_repo = ocm.OciOcmRepository(
                    type=ocm.OciAccess,
                    baseUrl=ocm_repo,
                )
            if not isinstance(ocm_repo, ocm.OciOcmRepository):
                raise NotImplementedError(ocm_repo)

            try:
                for version_tag in component_versions(
                    component_name=component_name,
                    ctx_repo=ocm_repo,
                    oci_client=oci_client,
                ):
                    versions.add(version_tag)
            except requests.exceptions.HTTPError as e:
                if (error_code := e.response.status_code) == 404:
                    continue

                image_reference = ocm_repo.component_oci_ref(component_name)
                if error_code == error_code_indicating_not_found(image_reference=image_reference):
                    continue

                raise

        if not versions and not absent_ok:
            raise om.OciImageNotFoundException()

        return versions

    return lookup


def composite_component_descriptor_lookup(
    lookups: tuple[ComponentDescriptorLookupById, ...],
    ocm_repository_lookup: OcmRepositoryLookup | None=None,
    default_absent_ok=True,
) -> ComponentDescriptorLookupById:
    '''
    Used to combine multiple ComponentDescriptorLookupByIds. The single lookups are used in
    the order they are specified. If the required component descriptor is found, it is
    written back to the prior lookups (if they have a WriteBack defined).

    @param lookups:
        a tuple of ComponentDescriptorLookupByIds which should be combined
    @param ocm_repository_lookup:
        ocm_repository_lookup to be used if none is specified in the lookup function
    @param default_absent_ok:
        sets the default behaviour in case of absent component descriptors for the returned lookup
        function
    '''
    def lookup(
        component_id: ocm.ComponentIdentity,
        /,
        ctx_repo: ocm.OciOcmRepository|str=None,
        ocm_repository_lookup=ocm_repository_lookup,
        absent_ok=default_absent_ok,
    ):
        component_id = cnudie.util.to_component_id(component_id)
        writebacks = []
        for lookup in lookups:
            res = None
            try:
                if ctx_repo:
                    res = lookup(
                        component_id,
                        ctx_repo=ctx_repo,
                    )
                else:
                    res = lookup(component_id)
            except om.OciImageNotFoundException:
                pass
            except dacite.DaciteError as ce:
                ce.add_note(f'{component_id=}')
                raise ce
            except requests.exceptions.HTTPError as he:
                if he.response.status_code != 500:
                    raise
                logger.warning(f'caught error {he} in {lookup=}, will try next lookup if any')

            if isinstance(res, ocm.ComponentDescriptor):
                for wb in writebacks: wb(component_id, res)
                return res
            elif res is None: continue
            elif isinstance(res, WriteBack): writebacks.append(res)

        # component descriptor not found in lookup
        if absent_ok:
            return

        if isinstance(ctx_repo, str):
            ctx_repo = ocm.OciOcmRepository(
                type=ocm.OciAccess,
                baseUrl=ctx_repo,
            )

        if ctx_repo:
            error = ctx_repo.component_version_oci_ref(component_id)
        elif ocm_repository_lookup:
            def to_repo_url(ocm_repo):
                if isinstance(ocm_repo, str):
                    return ocm_repo
                else:
                    return ocm_repo.oci_ref

            ocm_repository_urls = '\n'.join(
                to_repo_url(ocm_repository) for ocm_repository
                in ocm_repository_lookup(component_id)
            )
            error = f'Did not find {component_id=} in any of the following\n'
            error += f'ocm-repositories:\n{ocm_repository_urls}:\n{str(component_id)}'
        else:
            error = f'<no ocm-repo given>: {str(component_id)}'

        raise om.OciImageNotFoundException(
            error,
        )

    return lookup


def create_default_component_descriptor_lookup(
    ocm_repository_lookup: OcmRepositoryLookup=None,
    cache_dir: str | None=None,
    oci_client: oc.Client | collections.abc.Callable[[], oc.Client]=None,
    delivery_client=None,
    default_absent_ok: bool=False,
    fallback_to_service_mapping: bool=True,
) -> ComponentDescriptorLookupById:
    '''
    This is a convenience function combining commonly used/recommended lookups, using global
    configuration if available. It combines (in this order) an in-memory cache, file-system cache,
    delivery-service based, and oci-registry based lookup.

    @param ocm_repository_lookup:
        lookup for OCM repositories
    @param cache_dir:
        directory used for caching. If cache_dir is not specified, the filesystem cache lookup is
        not included in the returned lookup
    @param oci_client:
        client to establish the connection to the oci-registry. If the client cannot be created, a
        ValueError is raised
    @param delivery_client:
        client to establish the connection to the delivery-service. If the client cannot be created,
        the delivery-service based lookup is not included in the returned lookup
    @param default_absent_ok:
        sets the default behaviour in case of absent component descriptors for the returned lookup
        function
    @param fallback_to_service_mapping:
        if set, it is tried to retrieve the requested component descriptor using the OCM repository
        mapping of thedelivery-service, in case it could not be retrieved using
        `ocm_repository_lookup`
    '''
    if not ocm_repository_lookup:
        import ctx
        ocm_repository_lookup = ctx.cfg.ctx.ocm_repository_lookup

    lookups = [
        in_memory_cache_component_descriptor_lookup(
            ocm_repository_lookup=ocm_repository_lookup,
        )
    ]
    if not cache_dir:
        import ctx
        if ctx.cfg:
            cache_dir = ctx.cfg.ctx.cache_dir

    if cache_dir:
        lookups.append(
            file_system_cache_component_descriptor_lookup(
                cache_dir=cache_dir,
                ocm_repository_lookup=ocm_repository_lookup,
            )
        )

    if not delivery_client:
        import ccc.delivery
        delivery_client = ccc.delivery.default_client_if_available()
    if delivery_client:
        lookups.append(delivery_service_component_descriptor_lookup(
            delivery_client=delivery_client,
            ocm_repository_lookup=ocm_repository_lookup,
            fallback_to_service_mapping=fallback_to_service_mapping,
        ))

    lookups.append(
        oci_component_descriptor_lookup(
            oci_client=oci_client,
            ocm_repository_lookup=ocm_repository_lookup,
        ),
    )

    return composite_component_descriptor_lookup(
        lookups=tuple(lookups),
        ocm_repository_lookup=ocm_repository_lookup,
        default_absent_ok=default_absent_ok,
    )


def component_diff(
    left_component: ocm.Component | ocm.ComponentDescriptor,
    right_component: ocm.Component | ocm.ComponentDescriptor,
    ignore_component_names=(),
    component_descriptor_lookup: ComponentDescriptorLookupById=None,
) -> cnudie.util.ComponentDiff:
    import cnudie.iter as ci # late import to avoid cyclic dependencies

    left_component = cnudie.util.to_component(left_component)
    right_component = cnudie.util.to_component(right_component)

    if not component_descriptor_lookup:
        component_descriptor_lookup = create_default_component_descriptor_lookup()

    left_components = tuple(
        component_node.component for component_node in ci.iter(
            component=left_component,
            lookup=component_descriptor_lookup,
            node_filter=ci.Filter.components,
        ) if component_node.component.name not in ignore_component_names
    )
    right_components = tuple(
        component_node.component for component_node in ci.iter(
            component=right_component,
            lookup=component_descriptor_lookup,
            node_filter=ci.Filter.components,
        ) if component_node.component.name not in ignore_component_names
    )

    return cnudie.util.diff_components(
        left_components=left_components,
        right_components=right_components,
        ignore_component_names=ignore_component_names,
    )


def component_versions(
    component_name: str,
    ctx_repo: ocm.OcmRepository,
    oci_client: oc.Client=None,
) -> collections.abc.Sequence[str]:
    if not isinstance(ctx_repo, ocm.OciOcmRepository):
        raise NotImplementedError(ctx_repo)

    if not oci_client:
        import ccc.oci
        oci_client = ccc.oci.oci_client()

    ctx_repo: ocm.OciOcmRepository
    oci_ref = ctx_repo.component_oci_ref(component_name)

    return oci_client.tags(image_reference=oci_ref)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import deprecated

from . import cluster_domain_from_kubernetes_config
from model.base import (
    NamedModelElement,
    ModelDefaultsMixin,
)
from model.proxy import DockerImageConfig


class WebhookDispatcherConfig(NamedModelElement, ModelDefaultsMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._apply_defaults(self.raw)

    def _required_attributes(self):
        return {
            'concourse_config_names',
        }

    def concourse_config_names(self):
        return self.raw['concourse_config_names']


WHD_DEPLOYMENT_SUBDOMAIN_LABEL = 'webhooks'


class WebhookDispatcherDeploymentConfig(NamedModelElement):
    def _required_attributes(self):
        return {
            'whd_image',
            'ingress_config',
            'external_url',
            'secrets_server_config',
            'kubernetes_config',
            'webhook_dispatcher_config',
            'container_port',
            'logging_els_index',
            'job_mapping_name',
        }

    @deprecated.deprecated
    def image_reference(self):
        image_config = self.image_config()
        return image_config.image_reference()

    def external_url(self):
        return self.raw.get('external_url')

    def ingress_config(self):
        return self.raw.get('ingress_config')

    def secrets_server_config_name(self):
        return self.raw.get('secrets_server_config')

    def job_mapping_name(self) -> str:
        return self.raw.get('job_mapping_name')

    def ingress_host(self, cfg_factory):
        cluster_domain = cluster_domain_from_kubernetes_config(
            cfg_factory,
            self.kubernetes_config_name(),
        )
        return f'{WHD_DEPLOYMENT_SUBDOMAIN_LABEL}.{cluster_domain}'

    def events(self):
        return self.raw.get('events', ['*'])

    def kubernetes_config_name(self):
        return self.raw.get('kubernetes_config')

    def webhook_dispatcher_config_name(self):
        return self.raw.get('webhook_dispatcher_config')

    def webhook_dispatcher_container_port(self):
        return self.raw['container_port']

    def image_config(self):
        return DockerImageConfig(self.raw.get('whd_image'))

    def logging_els_index(self):
        '''Name of the elastic-search index to log into'''
        return self.raw['logging_els_index']
